Type: Package
Package: smoothr
Title: Smooth and Tidy Spatial Features
Version: 1.3.0
Authors@R: 
    person(given = "Matthew",
           family = "Strimas-Mackey",
           role = c("aut", "cre"),
           email = "mstrimas@gmail.com",
           comment = c(ORCID = "0000-0001-8929-7776"))
Description: Tools for smoothing and tidying spatial features
    (i.e. lines and polygons) to make them more aesthetically pleasing.
    Smooth curves, fill holes, and remove small fragments from lines and
    polygons.
License: GPL-3
URL: https://strimas.com/smoothr/, https://github.com/mstrimas/smoothr
BugReports: https://github.com/mstrimas/smoothr/issues
Depends: R (>= 4.1.0)
Imports: sf, stats, units
Suggests: knitr, lwgeom, methods, rmarkdown, sp, testthat, terra
VignetteBuilder: knitr
Encoding: UTF-8
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-04 13:50:49 UTC; mes335
Author: Matthew Strimas-Mackey [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8929-7776>)
Maintainer: Matthew Strimas-Mackey <mstrimas@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-04 15:00:08 UTC
Built: R 4.6.0; ; 2026-05-05 00:08:40 UTC; unix

Type: Package
Package: smoothr
Title: Smooth and Tidy Spatial Features
Version: 1.3.0
Authors@R: 
    person(given = "Matthew",
           family = "Strimas-Mackey",
           role = c("aut", "cre"),
           email = "mstrimas@gmail.com",
           comment = c(ORCID = "0000-0001-8929-7776"))
Description: Tools for smoothing and tidying spatial features
    (i.e. lines and polygons) to make them more aesthetically pleasing.
    Smooth curves, fill holes, and remove small fragments from lines and
    polygons.
License: GPL-3
URL: https://strimas.com/smoothr/, https://github.com/mstrimas/smoothr
BugReports: https://github.com/mstrimas/smoothr/issues
Depends: R (>= 4.1.0)
Imports: sf, stats, units
Suggests: knitr, lwgeom, methods, rmarkdown, sp, testthat, terra
VignetteBuilder: knitr
Encoding: UTF-8
LazyData: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-04 13:50:49 UTC; mes335
Author: Matthew Strimas-Mackey [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8929-7776>)
Maintainer: Matthew Strimas-Mackey <mstrimas@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-04 15:00:08 UTC
Built: R 4.4.3; ; 2026-05-04 21:44:14 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(drop_crumbs,SpatVector)
S3method(drop_crumbs,Spatial)
S3method(drop_crumbs,sf)
S3method(drop_crumbs,sfc)
S3method(fill_holes,SpatVector)
S3method(fill_holes,Spatial)
S3method(fill_holes,sf)
S3method(fill_holes,sfc)
S3method(smooth,SpatVector)
S3method(smooth,Spatial)
S3method(smooth,sf)
S3method(smooth,sfc)
S3method(smooth,sfg)
export(densify)
export(drop_crumbs)
export(fill_holes)
export(smooth)
export(smooth_chaikin)
export(smooth_densify)
export(smooth_ksmooth)
export(smooth_spline)

## ----setup, include = FALSE---------------------------------------------------
run <- tolower(Sys.info()[["sysname"]]) != "sunos"
knitr::opts_chunk$set(
  eval = run,
  collapse = TRUE,
  warning = FALSE,
  message = FALSE,
  error = FALSE,
  fig.align = "center",
  out.width = "100%",
  comment = "#>"
)

## ----packages-----------------------------------------------------------------
library(terra)
library(sf)
library(units)
library(smoothr)

## ----jagged-polygons, echo=FALSE----------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(4, 0, 0, 0), mfrow = c(3, 3))
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA)
}

## ----jagged-polygons_print, echo=FALSE----------------------------------------
print(jagged_polygons)

## ----jagged-lines, echo=FALSE-------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3)
}

## ----jagged-lines_print, echo=FALSE-------------------------------------------
print(jagged_lines)

## ----guass-field, results="hide", dev="png", echo=-1--------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
file <- system.file("extdata", "jagged-raster.tif", package = "smoothr")
jagged_raster <- rast(file)
plot(jagged_raster, col = heat.colors(100), axes = FALSE)

## ----chaikin-polygons, echo=-1------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
p_smooth_chaikin <- smooth(jagged_polygons, method = "chaikin")
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA)
  plot(st_geometry(p_smooth_chaikin[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
}

## ----chaikin-lines, echo=-1---------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
l_smooth_chaikin <- smooth(jagged_lines, method = "chaikin")
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3)
  plot(st_geometry(l_smooth_chaikin[i, ]), col = "#E41A1C", lwd = 2, add = TRUE)
}

## ----smooth-polygons, echo=-1-------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
p_smooth_ksmooth <- smooth(jagged_polygons, method = "ksmooth")
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA)
  plot(st_geometry(p_smooth_ksmooth[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
}

## ----ksmooth-lines, echo=-1---------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
l_smooth_ksmooth <- smooth(jagged_lines, method = "ksmooth")
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3)
  plot(st_geometry(l_smooth_ksmooth[i, ]), col = "#E41A1C", lwd = 2, add = TRUE)
}

## ----spline-polygons, echo=-1-------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
p_smooth_spline <- smooth(jagged_polygons, method = "spline")
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = NA)
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA,
       add = TRUE)
  plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
}

## ----spline-lines, echo=-1----------------------------------------------------
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
l_smooth_spline <- smooth(jagged_lines, method = "spline")
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(l_smooth_spline[i, ]), col = NA)
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3, add = TRUE)
  plot(st_geometry(l_smooth_spline[i, ]), col = "#E41A1C", lwd = 2, add = TRUE)
}

## ----densify-n, echo=-1-------------------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
l <- jagged_lines$geometry[[2]]
# split every segment into 2
l_dense <- densify(l, n = 2)
plot(l, lwd = 5)
plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19, add = TRUE)

## ----densify-md, echo=-1------------------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
l <- jagged_lines$geometry[[2]]
# split every segment into 2
l_dense <- densify(l, max_distance = 0.1)
plot(l, lwd = 5)
plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19, add = TRUE)

## ----drop-crumbs-polygons, echo=-1--------------------------------------------
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
p <- jagged_polygons$geometry[7]
area_thresh <- units::set_units(200, km^2)
p_dropped <- drop_crumbs(p, threshold = area_thresh)
plot(p, col = "black", main = "Original")
plot(p_dropped, col = "black", main = "After drop_crumbs()")

## ----drop-crumbs-lines, echo=-1-----------------------------------------------
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
l <- jagged_lines$geometry[8]
# note that any units can be used
# conversion to units of projection happens automatically
length_thresh <- units::set_units(25, miles)
l_dropped <- drop_crumbs(l, threshold = length_thresh)
plot(l, lwd = 5, main = "Original")
plot(l_dropped, lwd = 5, main = "After drop_crumbs()")

## ----fill-polygons, echo=-1---------------------------------------------------
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
p <- jagged_polygons$geometry[5]
area_thresh <- units::set_units(800, km^2)
p_dropped <- fill_holes(p, threshold = area_thresh)
# plot
plot(p, col = "black", main = "Original")
plot(p_dropped, col = "black", main = "After fill_holes()")

## ----polygonize, dev="png", echo=-1-------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
file <- system.file("extdata", "jagged-raster.tif", package = "smoothr")
jagged_raster <- rast(file)

# pres/abs map
r <- classify(jagged_raster, rcl = c(-Inf, 0.5, Inf)) |>
  as.numeric()
plot(r, col = c("white", "#4DAF4A"), legend = FALSE, axes = FALSE)

# polygonize
r_poly <- ifel(r > 0, r, NA) |>
  as.polygons() |>
  st_as_sf()
plot(r_poly, col = NA, border = "grey20", lwd = 1.5, add = TRUE)

## ----raster-drop, echo=-1-----------------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_dropped <- drop_crumbs(r_poly, set_units(101, km^2))
# plot
plot(r_poly_dropped, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)

## ----raster-fill, echo=-1-----------------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_filled <- fill_holes(r_poly_dropped, set_units(201, km^2))
# plot
plot(r_poly_filled, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)

## ----smooth-raster, echo=-1---------------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_smooth <- smooth(r_poly_filled, method = "ksmooth")
# plot
plot(r_poly_smooth, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)

## ----smooth-raster-more, echo=-1----------------------------------------------
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_smooth <- smooth(r_poly_filled, method = "ksmooth", smoothness = 2)
# plot
plot(r_poly_smooth, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)

#' Densify lines or polygons
#'
#' This function adds additional vertices to lines or polygons via linear
#' interpolation, always while keeping the original vertices. Each line segment
#' will be split into equal length sub-segments. This densification algorithm
#' treats all vertices as Euclidean points, i.e. new points will not fall on a
#' great circle between existing vertices, rather they'll be along a straight
#' line.
#'
#' This function works on matrices of points and is generally not called
#' directly. Instead, use [smooth()] with `method = "densify"` to apply this
#' smoothing algorithm to spatial features.
#'
#' @param x numeric matrix; matrix of coordinates.
#' @param wrap logical; whether the coordinates should be wrapped at the ends,
#'   as for polygons and closed lines, to ensure a smooth edge.
#' @param n integer; number of times to split each line segment. Ignored if
#'   `max_distance` is specified.
#' @param max_distance numeric; the maximum distance between vertices in the
#'   resulting matrix. This is the Euclidean distance and not the great circle
#'   distance.
#'
#' @return A matrix with the coordinates of the densified curve.
#' @export
#' @examples
#' # smooth_densify works on matrices of coordinates
#' # use the matrix of coordinates defining a line as an example
#' m <- jagged_lines$geometry[[2]][]
#' m_dense <- smooth_densify(m, n = 5)
#' class(m)
#' class(m_dense)
#' plot(m, type = "b", pch = 19, cex = 1.5, axes = FALSE, xlab = NA, ylab = NA)
#' points(m_dense, col = "red", pch = 19, cex = 0.5)
#'
#' # max_distance can be used to ensure vertices are at most a given dist apart
#' m_md <- smooth_densify(m, max_distance = 0.05)
#' plot(m, type = "b", pch = 19, cex = 1.5, axes = FALSE, xlab = NA, ylab = NA)
#' points(m_md, col = "red", pch = 19, cex = 0.5)
#'
#' # smooth is a wrapper for smooth_densify that works on spatial features
#' library(sf)
#' l <- jagged_lines$geometry[[2]]
#' l_dense <- smooth(l, method = "densify", n = 2)
#' class(l)
#' class(l_dense)
#' plot(l, lwd = 5)
#' plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
#' plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19,
#'      add = TRUE)
smooth_densify <- function(x, wrap = FALSE, n = 10L, max_distance) {
  stopifnot(is.matrix(x), nrow(x) > 1, ncol(x) > 1)

  n_pts <- nrow(x)
  # set densification parameters
  if (missing(max_distance)) {
    stopifnot(is_count(n), n >= 1)
    # n segments = n + 1 points
    # repeat for each segment
    n <- rep(n + 1, n_pts - 1)
  } else {
    stopifnot(
      is.numeric(max_distance),
      length(max_distance) == 1,
      max_distance > 0
    )
    # determine number of points based on max distance
    n <- ceiling(point_distance(x) / max_distance) + 1
  }

  # generate evenly spaced points for each dimension
  pts_dense <- NULL
  for (i in seq_len(ncol(x))) {
    sm <- seq_multiple(start = x[1:(n_pts - 1), i], end = x[2:n_pts, i], n = n)
    pts_dense <- cbind(pts_dense, sm)
  }
  colnames(pts_dense) <- colnames(x)

  # make sure start and end points are the same if wrapped
  if (wrap) {
    pts_dense[nrow(pts_dense), ] <- pts_dense[1, ]
  }
  pts_dense
}

context("smooth_densify()")
skip_on_os("solaris")

test_that("smooth_densify() works on polygons", {
  m <- jagged_polygons$geometry[[2]][[1]]
  m_smooth <- smooth_densify(m, wrap = TRUE, n = 2)
  expect_is(m_smooth, "matrix")
  expect_equal(nrow(m_smooth), 2 * (nrow(m) - 1) + 1)
  expect_equal(m_smooth[1, ], m_smooth[nrow(m_smooth), ])
})

test_that("smooth_densify() works on lines", {
  l <- jagged_lines$geometry[[2]][]
  l_smooth <- smooth_densify(l, wrap = FALSE, n = 3)
  expect_is(l_smooth, "matrix")
  expect_equal(nrow(l_smooth), 3 * (nrow(l) - 1) + 1)
})

test_that("smooth_densify() works on 3d lines", {
  g <- sf::st_geometry(jagged_lines_3d)
  for (i in seq_along(g)) {
    l <- g[[i]][]
    l_smooth <- smooth_densify(l, wrap = FALSE, n = 3)
    expect_is(l_smooth, "matrix")
    expect_equal(nrow(l_smooth), 3 * (nrow(l) - 1) + 1)
  }
})

test_that("smooth_densify() max_distance works", {
  l <- jagged_lines$geometry[[2]][]
  md <- 0.1
  l_smooth <- smooth_densify(l, wrap = FALSE, max_distance = md)
  expect_is(l_smooth, "matrix")
  expect_true(all(smoothr:::point_distance(l_smooth) <= md))

  m <- jagged_polygons$geometry[[2]][[1]]
  m_smooth <- smooth_densify(m, wrap = TRUE, max_distance = md)
  expect_is(m_smooth, "matrix")
  expect_true(all(smoothr:::point_distance(m_smooth) <= md))
  expect_equal(m_smooth[1, ], m_smooth[nrow(m_smooth), ])
})

test_that("smooth_densify() edge cases work", {
  m <- jagged_lines$geometry[[2]][]
  # if max distance is longer than segment distances, return original vertices
  expect_equivalent(m, smooth_densify(m, max_distance = 10000))
  # if n = 1 return original vertices
  expect_equivalent(m, smooth_densify(m, n = 1))
})

test_that("smooth_densify() raises error on invalid input", {
  expect_error(smooth_densify(jagged_lines))
  m <- jagged_lines$geometry[[2]][]
  expect_error(smooth_densify(m, n = -1))
  expect_error(smooth_densify(m, n = 0))
  expect_error(smooth_densify(m, n = 1.5))
  expect_error(smooth_densify(m, max_distance = -1.0))
  expect_error(smooth_densify(m, max_distance = 0))
})

#' Densify spatial lines or polygons
#'
#' A wrapper for `smooth(x, method = "densify")`. This function adds additional
#' vertices to spatial feature via linear interpolation, always while keeping
#' the original vertices. Each line segment will be split into equal length
#' sub-segments. This densification algorithm treats all vertices as Euclidean
#' points, i.e. new points will not fall on a great circle between existing
#' vertices, rather they'll be along a straight line.
#'
#' @param x spatial features; lines or polygons from either the `sf`, `sp`, or
#'   `terra` packages.
#' @inheritParams smooth_densify
#'
#' @return A densified polygon or line in the same format as the input data.
#' @export
#' @examples
#' library(sf)
#' l <- jagged_lines$geometry[[2]]
#' l_dense <- densify(l, n = 2)
#' plot(l, lwd = 5)
#' plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
#' plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19,
#'      add = TRUE)
densify <- function(x, n = 10L, max_distance) {
  if (missing(max_distance)) {
    smooth(x, method = "densify", n = n)
  } else {
    smooth(x, method = "densify", max_distance = max_distance)
  }
}

is_integer <- function(x) {
  is.integer(x) || (is.numeric(x) && all(x == as.integer(x)))
}

is_count <- function(x) {
  is_integer(x) && length(x) == 1 && x >= 0
}

is_flag <- function(x) {
  is.logical(x) && length(x) == 1 && !is.na(x)
}

point_distance <- function(x) {
  d <- diff(x)^2
  sqrt(rowSums(d))
}

seq_multiple <- function(start, end, n) {
  f <- function(x, y, z) {
    sq <- seq(from = x, to = y, length.out = z)
    # remove start since it will be duplicated by end of previous segment
    sq[-1]
  }
  # interpolate evenly between a series of points
  sq_mult <- mapply(f, start, end, n, SIMPLIFY = FALSE)
  # combine segments, need to add overall start point
  c(start[1], do.call(c, sq_mult))
}

context("smooth()")
skip_on_os("solaris")

library(sf)

test_that("smooth() methods work", {
  # polygons
  # chakin
  s <- smooth(jagged_polygons, method = "chaikin")
  # change precision to fix some floating point issues on windows
  s <- st_set_precision(s, 1e6)
  expect_true(all(st_is_valid(s)))
  # ksmooth
  s <- smooth(jagged_polygons[1:6, ], method = "ksmooth")
  # change precision to fix some floating point issues on windows
  s <- st_set_precision(s, 1e6)
  expect_true(all(st_is_valid(s)))
  # spline
  s <- smooth(jagged_polygons[1:6, ], method = "spline")
  # change precision to fix some floating point issues on windows
  s <- st_set_precision(s, 1e6)
  expect_true(all(st_is_valid(s)))
  # densify
  s <- smooth(jagged_polygons, method = "densify")
  # change precision to fix some floating point issues on windows
  s <- st_set_precision(s, 1e6)
  expect_true(all(st_is_valid(s)))

  # lines
  sl <- smooth(jagged_lines, method = "chaikin")
  expect_true(all(st_is_valid(sl)))
  sl <- smooth(jagged_lines, method = "ksmooth")
  expect_true(all(st_is_valid(sl)))
  sl <- smooth(jagged_lines, method = "spline")
  expect_true(all(st_is_valid(sl)))
  sl <- smooth(jagged_lines, method = "densify")
  expect_true(all(st_is_valid(sl)))

  # test parameters
  # chaikin
  p <- jagged_polygons$geometry[[3]]
  s_r2 <- smooth(p, method = "chaikin", refinements = 1)
  s_r5 <- smooth(p, method = "chaikin", refinements = 5)
  expect_lt(nrow(s_r2[[1]]), nrow(s_r5[[1]]))
  # ksmooth
  s_n2 <- smooth(p, method = "ksmooth", n = 2)
  s_n3 <- smooth(p, method = "ksmooth", n = 3)
  expect_lt(nrow(s_n2[[1]]), nrow(s_n3[[1]]))
  expect_lte(abs(nrow(s_n2[[1]]) - (2 * (nrow(p[[1]]) - 1) + 1)), 1)
  expect_lte(abs(nrow(s_n3[[1]]) - (3 * (nrow(p[[1]]) - 1) + 1)), 1)
  s_md1 <- smooth(p, method = "ksmooth", max_distance = 0.1)
  s_md05 <- smooth(p, method = "ksmooth", max_distance = 0.05)
  expect_lt(nrow(s_md1[[1]]), nrow(s_md05[[1]]))
  # spline
  s_n100 <- smooth(p, method = "spline", n = 100)
  s_n1000 <- smooth(p, method = "spline", n = 1000)
  expect_lt(nrow(s_n100[[1]]), nrow(s_n1000[[1]]))
  expect_equal(nrow(s_n100[[1]]), 100)
  expect_equal(nrow(s_n1000[[1]]), 1000)
  s_vf2 <- smooth(p, method = "spline", vertex_factor = 2)
  s_vf4 <- smooth(p, method = "spline", vertex_factor = 4)
  expect_equal(nrow(s_vf2[[1]]), nrow(p[[1]]) * 2)
  expect_equal(nrow(s_vf4[[1]]), nrow(p[[1]]) * 4)
  # densify
  s_n2 <- smooth(p, method = "densify", n = 2)
  s_n3 <- smooth(p, method = "densify", n = 3)
  expect_lt(nrow(s_n2[[1]]), nrow(s_n3[[1]]))
  expect_equal(nrow(s_n2[[1]]), 2 * (nrow(p[[1]]) - 1) + 1)
  expect_equal(nrow(s_n3[[1]]), 3 * (nrow(p[[1]]) - 1) + 1)
  s_md1 <- smooth(p, method = "densify", max_distance = 0.1)
  s_md05 <- smooth(p, method = "densify", max_distance = 0.05)
  expect_lt(nrow(s_md1[[1]]), nrow(s_md05[[1]]))
  expect_true(all(smoothr:::point_distance(s_md1[[1]]) <= 0.1))
})

test_that("smooth() works for different input formats", {
  s_sf <- smooth(jagged_polygons)
  s_sfc <- smooth(st_geometry(jagged_polygons))
  s_sfg <- smooth(st_geometry(jagged_polygons)[[1]])
  s_spdf <- smooth(as(jagged_polygons, "Spatial"))
  s_sp <- smooth(as(as(jagged_polygons, "Spatial"), "SpatialPolygons"))
  expect_s3_class(s_sf, "sf")
  expect_s3_class(s_sfc, "sfc")
  expect_s3_class(s_sfg, "POLYGON")
  expect_s4_class(s_spdf, "SpatialPolygonsDataFrame")
  expect_s4_class(s_sp, "SpatialPolygons")
  expect_equal(nrow(s_sf), length(s_sfc))
  expect_equal(nrow(s_sf), length(s_spdf))
  expect_equivalent(st_set_geometry(s_sf, NULL), s_spdf@data)
})

test_that("smooth() works for SpatVector objects", {
  skip_if_not_installed("terra")
  jp_terra <- terra::vect(jagged_polygons)
  s_terra <- expect_warning(smooth(jp_terra))
  expect_s4_class(s_terra, "SpatVector")
})

test_that("smooth() preserves holes", {
  p <- jagged_polygons$geometry[[5]]
  expect_true(st_is_valid(smooth(p, method = "chaikin")))
  expect_true(st_is_valid(smooth(p, method = "ksmooth")))
  expect_true(st_is_valid(smooth(p, method = "spline")))
  expect_true(st_is_valid(smooth(p, method = "densify")))
  expect_equal(length(p), length(smooth(p)))
})

test_that("smooth() preserves multipart features", {
  p <- jagged_polygons$geometry[[7]]
  expect_true(st_is_valid(smooth(p, method = "chaikin")))
  expect_true(st_is_valid(smooth(p, method = "ksmooth")))
  expect_true(st_is_valid(smooth(p, method = "densify")))
  expect_equal(length(p), length(smooth(p)))

  l <- jagged_lines$geometry[[8]]
  expect_true(st_is_valid(smooth(l, method = "chaikin")))
  expect_true(st_is_valid(smooth(l, method = "ksmooth")))
  expect_true(st_is_valid(smooth(l, method = "spline")))
  expect_true(st_is_valid(smooth(l, method = "densify")))
  expect_equal(length(l), length(smooth(l)))
})

test_that("smooth() fails for points", {
  point <- st_point(c(0, 0))
  expect_error(smooth(point))
  expect_error(smooth(st_sfc(point)))
  expect_error(smooth(as(st_sfc(point), "Spatial")))
})

#' Chaikin's corner cutting algorithm
#'
#' Chaikin's corner cutting algorithm smooths a curve by iteratively replacing
#' every point by two new points: one 1/4 of the way to the next point and one
#' 1/4 of the way to the previous point.
#'
#' This function works on matrices of points and is generally not called
#' directly. Instead, use [smooth()] with `method = "chaikin"` to apply this
#' smoothing algorithm to spatial features.
#'
#' @param x numeric matrix; 2-column matrix of coordinates.
#' @param wrap logical; whether the coordinates should be wrapped at the ends,
#'   as for polygons and closed lines, to ensure a smooth edge.
#' @param refinements integer; number of corner cutting iterations to apply.
#'
#' @return A matrix with the coordinates of the smoothed curve.
#' @references The original reference for Chaikin's corner cutting algorithm is:
#'
#'   - Chaikin, G. An algorithm for high speed curve generation.
#'     *Computer Graphics and Image Processing* 3 (1974), 346-349.
#'
#'   This implementation was inspired by the following StackOverflow answer:
#'
#'   - <https://stackoverflow.com/a/47255374/3591386>
#' @seealso [smooth()]
#' @export
#' @examples
#' # smooth_chaikin works on matrices of coordinates
#' # use the matrix of coordinates defining a polygon as an example
#' m <- jagged_polygons$geometry[[2]][[1]]
#' m_smooth <- smooth_chaikin(m, wrap = TRUE)
#' class(m)
#' class(m_smooth)
#' plot(m, type = "l", axes = FALSE, xlab = NA, ylab = NA)
#' lines(m_smooth, col = "red")
#'
#' # smooth is a wrapper for smooth_chaikin that works on spatial features
#' library(sf)
#' p <- jagged_polygons$geometry[[2]]
#' p_smooth <- smooth(p, method = "chaikin")
#' class(p)
#' class(p_smooth)
#' plot(p)
#' plot(p_smooth, border = "red", add = TRUE)
smooth_chaikin <- function(x, wrap = FALSE, refinements = 3L) {
  stopifnot(is.matrix(x), nrow(x) > 1, ncol(x) > 1)
  stopifnot(is_flag(wrap))
  stopifnot(is_count(refinements), refinements <= 10)

  # polygons and closed lines need to be wrapped
  if (wrap) {
    for (i in seq_len(refinements)) {
      n_pts <- nrow(x)
      qr <- matrix(NA_real_, nrow = 2 * (n_pts - 1) + 1, ncol = ncol(x))
      qr[seq(1, nrow(qr) - 1, by = 2), ] <- 0.75 * x[-n_pts, ] + 0.25 * x[-1, ]
      qr[seq(2, nrow(qr) - 1, by = 2), ] <- 0.75 * x[-1, ] + 0.25 * x[-n_pts, ]
      qr[nrow(qr), ] <- qr[1, ]
      x <- qr
    }
    # lines should have endpoints fixed
  } else {
    for (i in seq.int(refinements)) {
      n_pts <- nrow(x)
      qr <- matrix(NA_real_, nrow = 2 * (n_pts - 1), ncol = ncol(x))
      qr[seq(1, nrow(qr), by = 2), ] <- 0.75 * x[-n_pts, ] + 0.25 * x[-1, ]
      qr[seq(2, nrow(qr), by = 2), ] <- 0.75 * x[-1, ] + 0.25 * x[-n_pts, ]
      qr[1, ] <- x[1, ]
      qr[nrow(qr), ] <- x[nrow(x), ]
      x <- qr
    }
  }
  x
}

context("smooth_chaikin()")
skip_on_os("solaris")

test_that("smooth_chaikin() works", {
  m <- jagged_polygons$geometry[[2]][[1]]
  m_smooth <- smooth_chaikin(m, wrap = TRUE)
  expect_is(m_smooth, "matrix")
  expect_gt(nrow(m_smooth), nrow(m))
  expect_equal(m_smooth[1, ], m_smooth[nrow(m_smooth), ])
})

test_that("smooth_chaikin() works on lines", {
  l <- jagged_lines$geometry[[2]][]
  l_smooth <- smooth_chaikin(l, wrap = FALSE)
  expect_is(l_smooth, "matrix")
  expect_gt(nrow(l_smooth), nrow(l))
})

test_that("smooth_chaikin() works on 3d lines", {
  g <- sf::st_geometry(jagged_lines_3d)
  for (i in seq_along(g)) {
    l <- g[[i]][]
    l_smooth <- smooth_chaikin(l, wrap = FALSE)
    expect_is(l_smooth, "matrix")
    expect_gt(nrow(l_smooth), nrow(l))
  }
})

test_that("smooth_chaikin() refinements = 0 returns original matrix", {
  m <- jagged_polygons$geometry[[2]][[1]]
  expect_equivalent(m, smooth_chaikin(m, wrap = TRUE, refinements = 0))
})

test_that("smooth_chaikin() raises error on invalid input", {
  expect_error(smooth_chaikin(jagged_polygons))
  m <- jagged_polygons$geometry[[2]][[1]]
  expect_error(smooth_chaikin(m, refinements = 100L))
  expect_error(smooth_chaikin(m, refinements = -1))
  expect_error(smooth_chaikin(m, refinements = 1.5))
})

#' `smoothr`: Smooth and Tidy Spatial Features
#'
#' Tools for smoothing and tidying spatial features (i.e. lines and polygons) to
#' make them more aesthetically pleasing. Smooth curves, fill holes, and remove
#' small fragments from lines and polygons.
#'
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

#' Jagged polygons for smoothing
#'
#' Spatial polygons in [sf][sf::sf] format for smoothing. Most of these polygons
#' have been created by converting rasters to polygons and therefore consist
#' entirely of right angles. There are examples of polygons with holes and
#' multipart polygons.
#'
#' @format An [sf][sf::sf] object with 9 features and 3 attributes:
#'   - **`type`**: character; the geometry, i.e. "polygon" or "line".
#'   - **`hole`**: logical; whether the polygon has holes or not.
#'   - **`multipart`**: logical; whether the feature is single or multipart.
"jagged_polygons"

#' Jagged lines for smoothing
#'
#' Spatial lines in [sf][sf::sf] format for smoothing. There are examples of
#' lines forming a closed loop and multipart lines.
#'
#' @format An [sf][sf::sf] object with 9 features and 3 attributes:
#'   - **`type`**: character; the geometry, i.e. "polygon" or "line".
#'   - **`closed`**: logical; whether the line forms a closed loop or not.
#'   - **`multipart`**: logical; whether the feature is single or multipart.
"jagged_lines"

#' 3D jagged line with Z-dimension for smoothing
#'
#' Spatial lines in [sf][sf::sf]  format for smoothing in three dimensions.
#' There are examples of open and closed loops
#'
#' @format An [sf][sf::sf] object with 9 features and 3 attributes:
#'   - **`type`**: character; the geometry, i.e. "polygon" or "line".
#'   - **`closed`**: logical; whether the line forms a closed loop or not.
#'   - **`multipart`**: logical; whether the feature is single or multipart.
"jagged_lines_3d"

#' Remove small polygons or line segments
#'
#' Remove polygons or line segments below a given area or length threshold.
#'
#' For multipart features, the removal threshold is applied to the individual
#' components. This means that, in some cases, an entire feature may be removed,
#' while in other cases, only parts of the multipart feature will be removed.
#'
#' @inheritParams smooth
#' @param threshold an area or length threshold, below which features will be
#'   removed. Provided either as a `units` object (see [units::set_units()]), or
#'   a numeric threshold in the units of the coordinate reference system. If `x`
#'   is in unprojected coordinates, a numeric threshold is assumed to be in
#'   meters.
#' @param drop_empty logical; whether features with sizes below the given
#'   threshold should be removed (the default) or kept as empty geometries. Note
#'   that `sp` objects cannot store empty geometries, so this argument will be
#'   ignored and empty geometries will always be removed.
#'
#' @return A spatial feature, with small pieces removed, in the same format as
#'   the input data. If none of the features are larger than the threshold,
#'   `sf` inputs will return a geometry set with zero features, and `sp` inputs
#'   will return `NULL`.
#' @export
#' @examples
#' # remove polygons smaller than 200km2
#' p <- jagged_polygons$geometry[7]
#' area_thresh <- units::set_units(200, km^2)
#' p_dropped <- drop_crumbs(p, threshold = area_thresh)
#' # plot
#' par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
#' plot(p, col = "black", main = "Original")
#' if (length(p_dropped) > 0) {
#'   plot(p_dropped, col = "black", main = "After drop_crumbs()")
#' }
#'
#'
#' # remove lines less than 25 miles
#' l <- jagged_lines$geometry[8]
#' # note that any units can be used
#' # conversion to units of projection happens automatically
#' length_thresh <- units::set_units(25, miles)
#' l_dropped <- drop_crumbs(l, threshold = length_thresh)
#' # plot
#' par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
#' plot(l, lwd = 5, main = "Original")
#' if (length(l_dropped)) {
#'   plot(l_dropped, lwd = 5, main = "After drop_crumbs()")
#' }
#'
drop_crumbs <- function(x, threshold, drop_empty = TRUE) {
  UseMethod("drop_crumbs")
}

#' @export
drop_crumbs.sfc <- function(x, threshold, drop_empty = TRUE) {
  stopifnot(inherits(threshold, c("units", "numeric")), length(threshold) == 1)
  # check geometry types and get units of feature size
  if (all(sf::st_is(x, c("POLYGON", "MULTIPOLYGON")))) {
    size_fxn <- sf::st_area
    geo_type <- "POLYGON"
  } else if (all(sf::st_is(x, c("LINESTRING", "MULTILINESTRING")))) {
    size_fxn <- sf::st_length
    geo_type <- "LINESTRING"
  } else {
    stop(paste(
      "drop_crumbs() only works for line and polygon features",
      "and geometry types cannot be mixed."
    ))
  }

  thresh_nounits <- as.numeric(threshold)
  if (thresh_nounits < 0) {
    stop("threshold cannot be negative")
  }

  # convert threshold to crs units
  sz <- size_fxn(x[1])
  if (inherits(sz, "units")) {
    size_units <- units::set_units(1, units(sz), mode = "standard")
    threshold <- units::set_units(threshold, size_units, mode = "standard")
    stopifnot(threshold > units::set_units(0, size_units, mode = "standard"))
  }

  # assign a unique polygon id before breaking up multipart geometries
  x <- sf::st_as_sf(x, id = seq_along(x))
  all_ids <- x[["id"]]
  # split up multipart geometries
  x[["type"]] <- as.character(sf::st_geometry_type(x))
  x <- sf::st_cast(x, paste0("MULTI", geo_type), warn = FALSE)
  x <- sf::st_cast(x, geo_type, warn = FALSE)
  # apply threshold
  x <- x[size_fxn(x) > threshold, ]

  # add fully dropped features back in as empty geometries
  missing_ids <- setdiff(all_ids, x[["id"]])
  if (!drop_empty && length(missing_ids) > 0) {
    empty <- rep(
      sf::st_as_sfc("GEOMETRYCOLLECTION EMPTY", crs = sf::st_crs(x)),
      length(missing_ids)
    )
    empty <- sf::st_as_sf(empty, id = missing_ids, type = "GEOMETRYCOLLECTION")
    if (nrow(x) > 0) {
      x <- rbind(x, empty)
    } else {
      x <- empty
    }
    x <- x[order(x[["id"]]), ]
  }

  # re-group
  comb_fxn <- function(y) sf::st_cast(sf::st_combine(y), y[["type"]][1])
  x <- tapply(x, INDEX = x[["id"]], FUN = comb_fxn, simplify = FALSE)
  do.call(c, x)
}

#' @export
drop_crumbs.sf <- function(x, threshold, drop_empty = TRUE) {
  g <- drop_crumbs(
    sf::st_geometry(x),
    threshold = threshold,
    drop_empty = FALSE
  )
  sf::st_geometry(x) <- g
  if (drop_empty) {
    x <- x[!sf::st_is_empty(x), ]
  }
  x
}

#' @export
drop_crumbs.Spatial <- function(x, threshold, drop_empty = TRUE) {
  if (!requireNamespace("sp", quietly = TRUE)) {
    stop("Install the sp package to use drop_crumbs on sp features.")
  }
  # convert to sf object then back
  prj <- sp::proj4string(x)
  if (inherits(x, c("SpatialPolygonsDataFrame", "SpatialLinesDataFrame"))) {
    x_sf <- sf::st_as_sf(x)
  } else if (inherits(x, c("SpatialPolygons", "SpatialLines"))) {
    x_sf <- sf::st_as_sfc(x)
  } else {
    stop(paste("No drop_crumbs method for class", class(x)))
  }
  x_sf <- sf::st_set_crs(x_sf, prj)

  clean <- drop_crumbs(x_sf, threshold = threshold, drop_empty = TRUE)

  if (all(sf::st_is_empty(clean))) {
    return(NULL)
  }
  clean <- sf::as_Spatial(clean)
  sp::proj4string(clean) <- prj
  clean
}

#' @export
drop_crumbs.SpatVector <- function(x, threshold, drop_empty = TRUE) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("Install the terra package to process SpatVector features.")
  }
  warning(
    "SpatVector objects are converted to sf objects in smoothr. ",
    "This conversion may introduce errors and increase the time ",
    "required to perform smoothing."
  )
  clean <- drop_crumbs(
    sf::st_as_sf(x),
    threshold = threshold,
    drop_empty = TRUE
  )
  terra::vect(clean)
}

context("drop_crumbs()")
skip_on_os("solaris")

library(sf)
library(units)

test_that("drop_crumbs() works", {
  p <- jagged_polygons$geometry[7]
  area_thresh <- units::set_units(200, km^2)
  p_dropped <- drop_crumbs(p, threshold = area_thresh)
  expect_true(any(st_area(st_cast(p, "POLYGON")) < area_thresh))
  expect_true(all(st_area(st_cast(p_dropped, "POLYGON")) >= area_thresh))

  l <- jagged_lines$geometry[8]
  length_thresh <- units::set_units(40, km)
  l_dropped <- drop_crumbs(l, threshold = length_thresh)
  expect_true(any(st_length(st_cast(l, "LINESTRING")) < length_thresh))
  expect_true(all(st_length(st_cast(l_dropped, "LINESTRING")) >= length_thresh))
})

test_that("drop_crumbs() doesn't alter features with nothing dropped", {
  p <- jagged_polygons
  p_dropped <- drop_crumbs(p, threshold = units::set_units(0.1, m^2))
  expect_equivalent(p, p_dropped)

  l <- jagged_lines
  l_dropped <- drop_crumbs(l, threshold = units::set_units(0.1, m))
  expect_equivalent(l, l_dropped)
})

test_that("drop_crumbs() drop_empty works", {
  p <- jagged_polygons$geometry
  area_thresh <- units::set_units(200, km^2)
  p_dropped <- drop_crumbs(p, threshold = area_thresh, drop_empty = FALSE)
  expect_equal(length(p_dropped), length(p))
  expect_true(any(st_is_empty(p_dropped)))
  p_dropped <- drop_crumbs(p, threshold = area_thresh, drop_empty = TRUE)
  expect_lt(length(p_dropped), length(p))
  expect_true(all(!st_is_empty(p_dropped)))

  l <- jagged_lines$geometry
  length_thresh <- units::set_units(50, km)
  l_dropped <- drop_crumbs(l, threshold = length_thresh, drop_empty = FALSE)
  expect_equal(length(l_dropped), length(l))
  expect_true(any(st_is_empty(l_dropped)))
  l_dropped <- drop_crumbs(l, threshold = length_thresh, drop_empty = TRUE)
  expect_lt(length(l_dropped), length(l))
  expect_true(all(!st_is_empty(l_dropped)))
})

test_that("drop_crumbs() handling of empty results", {
  s_sf <- drop_crumbs(jagged_polygons, threshold = 2e20)
  s_sfc <- drop_crumbs(st_geometry(jagged_polygons), threshold = 2e20)
  jp_sp <- as(jagged_polygons, "Spatial")
  sp::proj4string(jp_sp) <- st_crs(jagged_polygons)$proj4string
  s_spdf <- drop_crumbs(jp_sp, threshold = 2e20)

  expect_equal(nrow(s_sf), 0)
  expect_equal(length(s_sfc), 0)
  expect_null(s_spdf)
})

test_that("drop_crumbs() works for different input formats", {
  s_sf <- drop_crumbs(jagged_polygons, threshold = 2e8)
  s_sfc <- drop_crumbs(st_geometry(jagged_polygons), threshold = 2e8)

  jp_spdf <- as_Spatial(jagged_polygons)
  sp::proj4string(jp_spdf) <- st_crs(jagged_polygons)$proj4string
  s_spdf <- drop_crumbs(jp_spdf, threshold = 2e8)

  jp_sp <- as(jp_spdf, "SpatialPolygons")
  sp::proj4string(jp_sp) <- st_crs(jagged_polygons)$proj4string
  s_sp <- drop_crumbs(jp_sp, threshold = 2e8)

  expect_s3_class(s_sf, "sf")
  expect_s3_class(s_sfc, "sfc")
  expect_s4_class(s_spdf, "SpatialPolygonsDataFrame")
  expect_s4_class(s_sp, "SpatialPolygons")
  expect_equal(nrow(s_sf), length(s_sfc))
  expect_equal(nrow(s_sf), length(s_spdf))
  expect_equivalent(st_set_geometry(s_sf, NULL), s_spdf@data)
})

test_that("drop_crumbs() works for SpatVector objects", {
  skip_if_not_installed("terra")
  jp <- jagged_polygons[7, ]
  jp_terra <- terra::vect(jp)
  s_terra <- expect_warning(
    drop_crumbs(jp_terra, threshold = units::set_units(200, km^2))
  )
  expect_s4_class(s_terra, "SpatVector")

  a_diff <- terra::expanse(jp_terra) - terra::expanse(s_terra)
  expect_gt(a_diff, 0)
})

test_that("drop_crumbs() fails for points", {
  point <- st_point(c(0, 0)) |>
    st_sfc()
  expect_error(drop_crumbs(point, threshold = 1))
  expect_error(drop_crumbs(as(st_sfc(point), "Spatial"), threshold = 1))
})

test_that("drop_crumbs() fails for mixed geometries", {
  mixed <- list(jagged_polygons$geometry[[1]], jagged_lines$geometry[[1]]) |>
    st_sfc(crs = 4326)
  expect_error(drop_crumbs(mixed, threshold = 1))
})

test_that("drop_crumbs() fails for invalid thresholds", {
  expect_error(drop_crumbs(jagged_polygons, threshold = -1))
  expect_error(drop_crumbs(jagged_polygons, threshold = 0))
  expect_error(drop_crumbs(jagged_polygons, threshold = set_units(1, km)))
})

#' Fill small holes in polygons
#'
#' Fill polygon holes that fall below a given area threshold.
#'
#' @inheritParams smooth
#' @param threshold an area threshold, below which holes will be removed.
#'   Provided either as a `units` object (see [units::set_units()]), or a
#'   numeric threshold in the units of the coordinate reference system. If `x`
#'   is in unprojected coordinates, a numeric threshold is assumed to be in
#'   square meters. A threshold of 0 will return the input polygons unchanged.
#'
#' @return A spatial feature, with holes filled, in the same format as the input
#'   data.
#' @export
#' @examples
#' # fill holes smaller than 1000km2
#' p <- jagged_polygons$geometry[5]
#' area_thresh <- units::set_units(1000, km^2)
#' p_dropped <- fill_holes(p, threshold = area_thresh)
#' # plot
#' par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
#' plot(p, col = "black", main = "Original")
#' plot(p_dropped, col = "black", main = "After fill_holes()")
fill_holes <- function(x, threshold) {
  UseMethod("fill_holes")
}

#' @export
fill_holes.sfc <- function(x, threshold) {
  stopifnot(inherits(threshold, c("units", "numeric")), length(threshold) == 1)
  # check geometry types and get units of feature size
  if (!all(sf::st_is(x, c("POLYGON", "MULTIPOLYGON")))) {
    stop("fill_holes() only works for polygon features.")
  }

  # zero threshold returns the input features unchanged
  thresh_nounits <- as.numeric(threshold)
  if (thresh_nounits == 0) {
    return(x)
  } else if (thresh_nounits < 0) {
    stop("threshold cannot be negative")
  }

  # convert threshold to crs units
  if (is.na(sf::st_crs(x))) {
    area_units <- units::set_units(1, "m2")
  } else {
    area_units <- units::set_units(
      1,
      units(sf::st_area(x[1])),
      mode = "standard"
    )
  }
  threshold <- units::set_units(threshold, area_units, mode = "standard")

  x_crs <- sf::st_crs(x)

  # loop over features, potentially multipart
  for (i in seq_along(x)) {
    # split up multipart geometries
    singles <- sf::st_cast(x[i], "POLYGON")
    # loop over single features
    for (j in seq_along(singles)) {
      # skip features with no holes
      if (length(singles[[j]]) > 1) {
        # area of holes
        a <- vapply(singles[[j]][-1], hole_area, 1, crs = x_crs)
        # assign units
        a <- units::set_units(a, area_units, mode = "standard")
        # remove holes not passing threshold
        singles[[j]] <- sf::st_polygon(singles[[j]][c(TRUE, a >= threshold)])
      }
    }
    # recombine
    if (length(singles) != 1) {
      singles <- sf::st_combine(singles)
    }
    x[[i]] <- singles[[1]]
  }
  # remove empty geometries
  sf::st_sfc(x)
}

#' @export
fill_holes.sf <- function(x, threshold) {
  sf::st_set_geometry(x, fill_holes(sf::st_geometry(x), threshold = threshold))
}

#' @export
fill_holes.Spatial <- function(x, threshold) {
  if (!requireNamespace("sp", quietly = TRUE)) {
    stop("Install the sp package to use fill_holes on sp features.")
  }
  # convert to sf object then back
  prj <- sp::proj4string(x)
  if (inherits(x, c("SpatialPolygonsDataFrame"))) {
    x_sf <- sf::st_as_sf(x)
  } else if (inherits(x, c("SpatialPolygons"))) {
    x_sf <- sf::st_as_sfc(x)
  } else {
    stop(paste("No fill_holes method for class", class(x)))
  }
  x_sf <- sf::st_set_crs(x_sf, prj)

  clean <- fill_holes(x_sf, threshold = threshold)

  if (all(sf::st_is_empty(clean))) {
    return(NULL)
  }
  clean <- sf::as_Spatial(clean)
  sp::proj4string(clean) <- prj
  clean
}

#' @export
fill_holes.SpatVector <- function(x, threshold) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("Install the terra package to process SpatVector features.")
  }
  warning(
    "SpatVector objects are converted to sf objects in smoothr. ",
    "This conversion may introduce errors and increase the time ",
    "required to perform smoothing."
  )

  # convert to sf object then back
  clean <- fill_holes(sf::st_as_sf(x), threshold = threshold)
  terra::vect(clean)
}

hole_area <- function(x, crs) {
  sf::st_area(sf::st_sfc(sf::st_polygon(list(x)), crs = crs))
}

context("fill_holes()")
skip_on_os("solaris")

library(sf)
library(units)

test_that("fill_holes() works", {
  p <- jagged_polygons$geometry[5]
  # remove hole
  p_filled <- fill_holes(p, threshold = units::set_units(1000, km^2))
  expect_true(st_area(p) < st_area(p_filled))
  # don't remove hole
  p_filled <- fill_holes(p, threshold = units::set_units(100, km^2))
  expect_true(st_area(p) == st_area(p_filled))
})

test_that("fill_holes() doesn't alter polygons with no holes", {
  p <- jagged_polygons[!jagged_polygons$hole, ]
  p_filled <- fill_holes(p, threshold = units::set_units(1e12, km^2))
  expect_equivalent(p, p_filled)
})

test_that("fill_holes() works for different input formats", {
  jp <- jagged_polygons[5:6, ]
  s_sf <- fill_holes(jp, threshold = 2e8)
  s_sfc <- fill_holes(st_geometry(jp), threshold = 2e8)

  jp_sp <- as(jp, "Spatial")
  sp::proj4string(jp_sp) <- st_crs(jp)$proj4string

  s_spdf <- fill_holes(jp_sp, threshold = 2e8)
  s_sp <- fill_holes(as(jp_sp, "SpatialPolygons"), threshold = 2e8)
  expect_s3_class(s_sf, "sf")
  expect_s3_class(s_sfc, "sfc")
  expect_s4_class(s_spdf, "SpatialPolygonsDataFrame")
  expect_s4_class(s_sp, "SpatialPolygons")
  a_diff <- st_area(s_sf) - st_area(s_sfc)
  expect_true(sum(abs(a_diff)) < units::set_units(1, "m^2"))
  a_diff <- st_area(s_sf) - st_area(st_as_sf(s_spdf))
  expect_true(sum(abs(a_diff)) < units::set_units(1, "m^2"))
  expect_equivalent(st_set_geometry(s_sf, NULL), s_spdf@data)
})

test_that("fill_holes() works for SpatVector objects", {
  skip_if_not_installed("terra")
  jp <- jagged_polygons[5:6, ]
  jp_terra <- terra::vect(jp)
  s_terra <- expect_warning(
    fill_holes(jp_terra, threshold = units::set_units(1000, km^2))
  )
  expect_s4_class(s_terra, "SpatVector")

  a_diff <- terra::expanse(jp_terra) - terra::expanse(s_terra)
  expect_lt(a_diff[1], 0)
  expect_equal(a_diff[2], 0)
})

test_that("fill_holes() fails for points and lines", {
  point <- st_point(c(0, 0)) |>
    st_sfc()
  expect_error(fill_holes(point, threshold = 1))
  expect_error(fill_holes(as(st_sfc(point), "Spatial"), threshold = 1))

  expect_error(fill_holes(jagged_lines, threshold = 1))
  expect_error(fill_holes(as(jagged_lines, "Spatial"), threshold = 1))
})

test_that("fill_holes() fails for invalid thresholds", {
  expect_error(fill_holes(jagged_polygons, threshold = -1))
  expect_error(fill_holes(jagged_polygons, threshold = set_units(1, "km")))
})

test_that("fill_holes() does nothing when threshold = 0", {
  p_filled <- fill_holes(jagged_polygons, threshold = 0)
  expect_identical(jagged_polygons, p_filled)
})

#' Spline interpolation
#'
#' Spline interpolation uses [stats::spline()] to interpolate between existing
#' vertices using piecewise cubic polynomials. The coordinates are interpolated
#' independently. The curve will always pass through the vertices of the
#' original feature.
#'
#' This function works on matrices of points and is generally not called
#' directly. Instead, use [smooth()] with `method = "spline"` to apply this
#' smoothing algorithm to spatial features.
#'
#' @param x numeric matrix; matrix of coordinates.
#' @param wrap logical; whether the coordinates should be wrapped at the ends,
#'   as for polygons and closed lines, to ensure a smooth edge.
#' @param vertex_factor double; the proportional increase in the number of
#'   vertices in the smooth curve. For example, if the original curve has 100
#'   points, a value of `2.5` will yield a new smoothed curve with 250 points.
#'   Ignored if `n` is specified.
#' @param n integer; number of vertices in the smoothed curve.
#'
#' @return A matrix with the coordinates of the smoothed curve.
#' @references The spline method was inspired by the following StackExchange
#'   answers:
#'
#'   - <https://stackoverflow.com/a/26089377/3591386>
#'   - <https://gis.stackexchange.com/a/24929/26661>
#'
#' @seealso [smooth()]
#' @export
#' @examples
#' # smooth_spline works on matrices of coordinates
#' # use the matrix of coordinates defining a polygon as an example
#' m <- jagged_polygons$geometry[[2]][[1]]
#' m_smooth <- smooth_spline(m, wrap = TRUE)
#' class(m)
#' class(m_smooth)
#' plot(m_smooth, type = "l", col = "red", axes = FALSE, xlab = NA, ylab = NA)
#' lines(m, col = "black")
#'
#' # smooth is a wrapper for smooth_spline that works on spatial features
#' library(sf)
#' p <- jagged_polygons$geometry[[2]]
#' p_smooth <- smooth(p, method = "spline")
#' class(p)
#' class(p_smooth)
#' plot(p_smooth, border = "red")
#' plot(p, add = TRUE)
smooth_spline <- function(x, wrap = FALSE, vertex_factor = 5, n) {
  stopifnot(is.matrix(x), nrow(x) > 1, ncol(x) > 1)

  n_pts <- nrow(x)
  stopifnot(is_flag(wrap))
  if (missing(n)) {
    stopifnot(
      is.numeric(vertex_factor),
      length(vertex_factor) == 1,
      vertex_factor >= 1
    )
    n <- max(round(vertex_factor * n_pts), n_pts)
  } else {
    stopifnot(is_count(n), n >= n_pts)
  }
  if (wrap) {
    method <- "periodic"
  } else {
    method <- "fmm"
  }
  pts_smooth <- NULL
  for (i in seq_len(ncol(x))) {
    xs <- stats::spline(seq_len(n_pts), x[, i], n = n, method = method)$y
    pts_smooth <- cbind(pts_smooth, xs)
  }
  pts_smooth
}

#' Smooth a spatial feature
#'
#' Smooth out the jagged or sharp corners of spatial lines or polygons to make
#' them appear more aesthetically pleasing and natural.
#'
#' @param x spatial features; lines or polygons from either the `sf`, `sp`, or
#'   `terra` packages.
#' @param method character; specifies the type of smoothing method to use.
#'   Possible methods are: `"chaikin"`, `"ksmooth"`, `"spline"`, and
#'   `"densify"`. Each method has one or more parameters specifying the amount
#'   of smoothing to perform. See Details for descriptions.
#' @param ... additional arguments specifying the amount of smoothing, passed on
#'   to the specific smoothing function, see Details below.
#'
#' @details Specifying a method calls one of the following underlying smoothing
#'   functions. Each smoothing method has one or more parameters that specify
#'   the extent of smoothing. Note that for multiple features, or multipart
#'   features, these parameters apply to each individual, singlepart feature.
#'
#'   - [smooth_chaikin()]: Chaikin's corner cutting algorithm smooths a curve by
#'   iteratively replacing every point by two new points: one 1/4 of the way to
#'   the next point and one 1/4 of the way to the previous point. Smoothing
#'   parameters:
#'     - `refinements`: number of corner cutting iterations to apply.
#'   - [smooth_ksmooth()]: kernel smoothing via the [stats::ksmooth()] function.
#'   This method first calls [smooth_densify()] to densify the feature, then
#'   applies Gaussian kernel regression to smooth the resulting points.
#'   Smoothing parameters:
#'     - `smoothness`: a positive number controlling the smoothness and level of
#'     generalization. At the default value of 1, the  bandwidth is chosen as
#'     the mean distance between adjacent vertices. Values greater than 1
#'     increase the bandwidth, yielding more highly smoothed and generalized
#'     features, and values less than 1 decrease the bandwidth, yielding less
#'     smoothed and generalized features.
#'     - `bandwidth`: the bandwidth of the Guassian kernel. If this argument is
#'     supplied, then `smoothness` is ignored and an optimal bandwidth is not
#'     estimated.
#'     - `n`: number of times to split each line segment in the densification
#'     step. Ignored if `max_distance` is specified.
#'     - `max_distance`: the maximum distance between vertices in the resulting
#'     features for the densification step. This is the Euclidean distance and
#'     not the great circle distance.
#'   - [smooth_spline()]: spline interpolation via the [stats::spline()]
#'   function. This method interpolates between existing vertices and can be
#'   used when the resulting smoothed feature should pass through the vertices
#'   of the input feature. Smoothing parameters:
#'     - `vertex_factor`: the proportional increase in the number of vertices in
#'     the smooth feature. For example, if the original feature has 100
#'     vertices, a value of 2.5 will yield a new, smoothed feature with 250
#'     vertices.  Ignored if `n` is specified.
#'     - `n`: number of vertices in each smoothed feature.
#'   - [smooth_densify()]: densification of vertices for lines and polygons.
#'   This is not a true smoothing algorithm, rather new vertices are added to
#'   each line segment via linear interpolation. Densification parameters:
#'     - `n`: number of times to split each line segment. Ignored if
#'     `max_distance` is specified.
#'     - `max_distance`: the maximum distance between vertices in the resulting
#'     feature. This is the Euclidean distance and not the great circle
#'     distance.
#'
#' @return A smoothed polygon or line in the same format as the input data.
#' @references See specific smoothing function help pages for references.
#' @seealso [smooth_chaikin()] [smooth_ksmooth()] [smooth_spline()]
#'   [smooth_densify()]
#' @export
#' @examples
#' library(sf)
#' # compare different smoothing methods
#' # polygons
#' par(mar = c(0, 0, 0, 0), oma = c(4, 0, 0, 0), mfrow = c(3, 3))
#' p_smooth_chaikin <- smooth(jagged_polygons, method = "chaikin")
#' p_smooth_ksmooth <- smooth(jagged_polygons, method = "ksmooth")
#' p_smooth_spline <- smooth(jagged_polygons, method = "spline")
#' for (i in 1:nrow(jagged_polygons)) {
#'   plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = NA)
#'   plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA,
#'        add = TRUE)
#'   plot(st_geometry(p_smooth_chaikin[i, ]), col = NA, border = "#E41A1C",
#'        lwd = 2, add = TRUE)
#'   plot(st_geometry(p_smooth_ksmooth[i, ]), col = NA, border = "#4DAF4A",
#'        lwd = 2, add = TRUE)
#'   plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = "#377EB8",
#'        lwd = 2, add = TRUE)
#' }
#' par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), new = TRUE)
#' plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n", axes = FALSE)
#' legend("bottom", legend = c("chaikin", "ksmooth", "spline"),
#'        col = c("#E41A1C", "#4DAF4A", "#377EB8"),
#'        lwd = 2, cex = 2, box.lwd = 0, inset = 0, horiz = TRUE)
#'
#' # lines
#' par(mar = c(0, 0, 0, 0), oma = c(4, 0, 0, 0), mfrow = c(3, 3))
#' l_smooth_chaikin <- smooth(jagged_lines, method = "chaikin")
#' l_smooth_ksmooth <- smooth(jagged_lines, method = "ksmooth")
#' l_smooth_spline <- smooth(jagged_lines, method = "spline")
#' for (i in 1:nrow(jagged_lines)) {
#'   plot(st_geometry(l_smooth_spline[i, ]), col = NA)
#'   plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3, add = TRUE)
#'   plot(st_geometry(l_smooth_chaikin[i, ]), col = "#E41A1C", lwd = 2,
#'        lty = 2, add = TRUE)
#'   plot(st_geometry(l_smooth_ksmooth[i, ]), col = "#4DAF4A", lwd = 2,
#'        lty = 2, add = TRUE)
#'   plot(st_geometry(l_smooth_spline[i, ]), col = "#377EB8", lwd = 2,
#'        lty = 2, add = TRUE)
#' }
#' par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), new = TRUE)
#' plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n", axes = FALSE)
#' legend("bottom", legend = c("chaikin", "ksmooth", "spline"),
#'        col = c("#E41A1C", "#4DAF4A", "#377EB8"),
#'        lwd = 2, cex = 2, box.lwd = 0, inset = 0, horiz = TRUE)
smooth <- function(
  x,
  method = c("chaikin", "ksmooth", "spline", "densify"),
  ...
) {
  UseMethod("smooth")
}

#' @export
smooth.sfg <- function(
  x,
  method = c("chaikin", "ksmooth", "spline", "densify"),
  ...
) {
  method <- match.arg(method)

  # choose smoother
  if (method == "chaikin") {
    smoother <- smooth_chaikin
  } else if (method == "ksmooth") {
    smoother <- smooth_ksmooth
  } else if (method == "spline") {
    smoother <- smooth_spline
  } else if (method == "densify") {
    smoother <- smooth_densify
  } else {
    stop(paste("Invalid smoothing method:", method))
  }

  # perform smoothing
  if (sf::st_is(x, "LINESTRING")) {
    m <- x[]
    x <- sf::st_linestring(smoother(m, wrap = all(m[1, ] == m[nrow(m), ]), ...))
  } else if (sf::st_is(x, "POLYGON")) {
    for (i in seq_along(x)) {
      x[[i]] <- smoother(x[[i]], wrap = TRUE, ...)
    }
  } else if (sf::st_is(x, "MULTILINESTRING")) {
    for (i in seq_along(x)) {
      m <- x[[i]]
      x[[i]] <- smoother(x[[i]], wrap = all(m[1, ] == m[nrow(m), ]), ...)
    }
  } else if (sf::st_is(x, "MULTIPOLYGON")) {
    for (i in seq_along(x)) {
      for (j in seq_along(x[[i]])) {
        x[[i]][[j]] <- smoother(x[[i]][[j]], wrap = TRUE, ...)
      }
    }
  } else {
    stop("smooth only valid for line and polygon features")
  }
  x
}

#' @export
smooth.sfc <- function(
  x,
  method = c("chaikin", "ksmooth", "spline", "densify"),
  ...
) {
  method <- match.arg(method)

  for (i in seq_along(x)) {
    x[[i]] <- smooth(x[[i]], method = method, ...)
  }
  sf::st_sfc(x)
}

#' @export
smooth.sf <- function(
  x,
  method = c("chaikin", "ksmooth", "spline", "densify"),
  ...
) {
  method <- match.arg(method)
  sf::st_geometry(x) <- smooth(sf::st_geometry(x), method = method, ...)
  x
}

#' @export
smooth.Spatial <- function(
  x,
  method = c("chaikin", "ksmooth", "spline", "densify"),
  ...
) {
  if (!requireNamespace("sp", quietly = TRUE)) {
    stop("Install the sp package to smooth sp features.")
  }
  method <- match.arg(method)
  # convert to sf object then back
  prj <- sp::proj4string(x)
  if (inherits(x, c("SpatialPolygonsDataFrame", "SpatialLinesDataFrame"))) {
    x_sf <- sf::st_as_sf(x)
  } else if (inherits(x, c("SpatialPolygons", "SpatialLines"))) {
    x_sf <- sf::st_as_sfc(x)
  } else {
    stop(paste("No smooth method for class", class(x)))
  }
  x_sf <- sf::st_set_crs(x_sf, prj)

  smoothed <- smooth(x_sf, method = method, ...)

  smoothed <- sf::as_Spatial(smoothed)
  sp::proj4string(smoothed) <- prj
  smoothed
}

#' @export
smooth.SpatVector <- function(
  x,
  method = c("chaikin", "ksmooth", "spline", "densify"),
  ...
) {
  if (!requireNamespace("terra", quietly = TRUE)) {
    stop("Install the terra package to smooth SpatVector features.")
  }
  warning(
    "SpatVector objects are converted to sf objects in smoothr. ",
    "This conversion may introduce errors and increase the time ",
    "required to perform smoothing."
  )
  method <- match.arg(method)
  # convert to sp object then back
  smoothed <- smooth(sf::st_as_sf(x), method = method, ...)
  terra::vect(smoothed)
}

#' Kernel smooth
#'
#' Kernel smoothing uses [stats::ksmooth()] to smooth out existing vertices
#' using Gaussian kernel regression. Kernel smoothing is applied to the `x` and
#' `y` coordinates are independently. Prior to smoothing, [smooth_densify()] is
#' called to generate additional vertices, and the smoothing is applied to this
#' densified set of vertices.
#'
#' Kernel smoothing both smooths and generalizes curves, and the extent of these
#' effects is dependent on the bandwidth of the smoothing kernel. Therefore,
#' choosing a sensible bandwidth is critical when using this method. The choice
#' of bandwidth will be dependent on the projection, scale, and desired amount
#' of smoothing and generalization. The are two methods of adjusting the
#' bandwidth. By default, the bandwidth will be set to the average distances
#' between adjacent vertices. The `smoothness` factor can then be used to adjust
#' this calculated bandwidth, values greater than 1 will lead to more smoothing,
#' values less than 1 will lead to less smoothing. Alternatively, the bandwidth
#' can be chosen manually with the `bandwidth` argument. Typically, users will
#' need to explore a range of bandwidths to determine which yields the best
#' results for their situation.
#'
#' This function works on matrices of points and is generally not called
#' directly. Instead, use [smooth()] with `method = "ksmooth"` to apply this
#' smoothing algorithm to spatial features.
#'
#' @param x numeric matrix; 2-column matrix of coordinates.
#' @param wrap logical; whether the coordinates should be wrapped at the ends,
#'   as for polygons and closed lines, to ensure a smooth edge.
#' @param smoothness numeric; a parameter controlling the bandwidth of the
#'   Gaussian kernel, and therefore the smoothness and level of generalization.
#'   By default, the bandwidth is chosen as the mean distance between adjacent
#'   points. The `smoothness` parameter is a multiplier of this chosen
#'   bandwidth, with values greater than 1 yielding more highly smoothed and
#'   generalized features and values less than 1 yielding less smoothed and
#'   generalized features.
#' @param bandwidth numeric; the bandwidth of the Guassian kernel. If this
#'   argument is supplied, then `smoothness` is ignored and an optimal bandwidth
#'   is not estimated.
#' @param n integer; number of times to split each line segment for
#'   [smooth_densify()]. Ignored if `max_distance` is specified.
#' @param max_distance numeric; the maximum distance between vertices for
#'   [smooth_densify()]. This is the Euclidean distance and not the great circle
#'   distance.
#'
#' @return A matrix with the coordinates of the smoothed curve.
#' @references The kernel smoothing method was inspired by the following
#'   StackExchange answers:
#'
#'   - <https://stats.stackexchange.com/a/143608/44268>
#'   - <https://gis.stackexchange.com/a/24929/26661>
#' @seealso [smooth()]
#' @export
#' @examples
#' # smooth_ksmooth works on matrices of coordinates
#' # use the matrix of coordinates defining a polygon as an example
#' m <- jagged_polygons$geometry[[2]][[1]]
#' m_smooth <- smooth_ksmooth(m, wrap = TRUE)
#' class(m)
#' class(m_smooth)
#' plot(m, type = "l", col = "black", lwd = 3, axes = FALSE, xlab = NA,
#'      ylab = NA)
#' lines(m_smooth, lwd = 3, col = "red")
#'
#' # lines can also be smoothed
#' l <- jagged_lines$geometry[[2]][]
#' l_smooth <- smooth_ksmooth(l, wrap = FALSE, max_distance = 0.05)
#' plot(l, type = "l", col = "black", lwd = 3, axes = FALSE, xlab = NA,
#'      ylab = NA)
#' lines(l_smooth, lwd = 3, col = "red")
#'
#' # explore different levels of smoothness
#' p <- jagged_polygons$geometry[[2]][[1]]
#' ps1 <- smooth_ksmooth(p, wrap = TRUE, max_distance = 0.01, smoothness = 0.5)
#' ps2 <- smooth_ksmooth(p, wrap = TRUE, max_distance = 0.01, smoothness = 1)
#' ps3 <- smooth_ksmooth(p, wrap = TRUE, max_distance = 0.01, smoothness = 2)
#' # plot
#' par(mar = c(0, 0, 0, 0), oma = c(10, 0, 0, 0))
#' plot(p, type = "l", col = "black", lwd = 3, axes = FALSE, xlab = NA,
#'      ylab = NA)
#' lines(ps1, lwd = 3, col = "#E41A1C")
#' lines(ps2, lwd = 3, col = "#4DAF4A")
#' lines(ps3, lwd = 3, col = "#377EB8")
#' par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), new = TRUE)
#' plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n", axes = FALSE)
#' legend("bottom", legend = c("0.5", "1", "2"),
#'        col = c("#E41A1C", "#4DAF4A", "#377EB8"),
#'        lwd = 3, cex = 2, box.lwd = 0, inset = 0, horiz = TRUE)
#'
#' library(sf)
#' p <- jagged_polygons$geometry[[2]]
#' p_smooth <- smooth(p, method = "ksmooth")
#' class(p)
#' class(p_smooth)
#' plot(p_smooth, border = "red")
#' plot(p, add = TRUE)
smooth_ksmooth <- function(
  x,
  wrap = FALSE,
  smoothness = 1,
  bandwidth,
  n = 10L,
  max_distance
) {
  stopifnot(is.matrix(x), nrow(x) > 1, ncol(x) > 1)
  stopifnot(is_flag(wrap))
  stopifnot(is.numeric(smoothness), length(smoothness) == 1, smoothness > 0)

  # estimate bandwidth
  if (missing(bandwidth)) {
    d_orig <- point_distance(x)
    bandwidth <- smoothness * mean(d_orig)
  }
  stopifnot(is.numeric(bandwidth), length(bandwidth) == 1, bandwidth > 0)

  # grab ends for padding
  if (!isTRUE(wrap)) {
    # extend start and end by one "edge"
    pad <- list(
      start = rbind(x[1, ], 2 * x[1, ] - x[2, ]),
      end = rbind(x[nrow(x), ], 2 * x[nrow(x), ] - x[nrow(x) - 1, ])
    )
    pad$start <- smooth_densify(
      pad$start[2:1, ],
      wrap = FALSE,
      n = n,
      max_distance = max_distance
    )
    pad$end <- smooth_densify(
      pad$end,
      wrap = FALSE,
      n = n,
      max_distance = max_distance
    )
  }

  # first densify
  x <- smooth_densify(x, wrap = wrap, n = n, max_distance = max_distance)
  n_pts <- nrow(x)

  if (isTRUE(wrap)) {
    # wrap vertices
    x <- rbind(x[-n_pts, ], x, x[-1, ])
    d <- c(0, cumsum(point_distance(x)))

    # smooth
    # parameterize x and y as functions of distance along curve
    pts_smooth <- NULL
    for (i in seq_len(ncol(x))) {
      ks <- stats::ksmooth(
        d,
        x[, i],
        n.points = length(d),
        kernel = "normal",
        bandwidth = bandwidth
      )
      pts_smooth <- cbind(pts_smooth, ks[["y"]])
      if (i == 1) {
        keep_rows <- (ks$x >= d[n_pts]) & (ks$x <= d[(2 * n_pts - 1)])
      }
    }

    # remove extra wrapped vertices
    pts_smooth <- pts_smooth[keep_rows, ]

    # ensure the endpoints match
    pts_smooth[nrow(pts_smooth), ] <- pts_smooth[1, ]
  } else {
    # pad ends to ensure smoothed line doesn't end early
    x <- rbind(pad$start, x, pad$end)
    d <- c(0, cumsum(point_distance(x)))

    # smooth
    pts_smooth <- NULL
    for (i in seq_len(ncol(x))) {
      ks <- stats::ksmooth(
        d,
        x[, i],
        n.points = length(d),
        kernel = "normal",
        bandwidth = bandwidth
      )
      pts_smooth <- cbind(pts_smooth, ks[["y"]])
      if (i == 1) {
        keep_rows <- (ks$x >= d[nrow(pad$start) + 1]) &
          (ks$x <= d[nrow(pad$start) + n_pts])
      }
    }

    # removing padding
    pts_smooth <- pts_smooth[keep_rows, ]

    # ensure endpoints are fixed
    pts_smooth[1, ] <- pad$start[nrow(pad$start), ]
    pts_smooth[nrow(pts_smooth), ] <- pad$end[1, ]
  }
  pts_smooth
}

---
title: "smoothr: Smooth and Tidy Spatial Features in R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteEncoding{UTF-8}
  %\VignetteIndexEntry{smoothr: Smooth and Tidy Spatial Features in R}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include = FALSE}
run <- tolower(Sys.info()[["sysname"]]) != "sunos"
knitr::opts_chunk$set(
  eval = run,
  collapse = TRUE,
  warning = FALSE,
  message = FALSE,
  error = FALSE,
  fig.align = "center",
  out.width = "100%",
  comment = "#>"
)
```

`smoothr` offers a variety of tools for smoothing and tidying spatial features (i.e. polygons and lines) to make them more aesthetically pleasing, especially when converting raster data to vector format. This package offers support for both `sp`, `sf`, and `terra` vector objects. The following smoothing methods are available:

- **Chaikin's corner cutting algorithm**: smoothing using Chaikin's corner cutting algorithm, which iteratively replaces every point by two new points: one 1/4 of the way to the next point and one 1/4 of the way to the previous point. This method applies a moderate amount of smoothing of sharp corners without extensive generalization.
- **Kernel smoothing:** smoothing using Gaussian kernel regression via the `ksmooth()` function. This approach first densifies the feature (i.e. adds more vertices) then applies the kernel smoothing. Kernel smoothing simultaneously smooths and generalizes curves, and can be turned to produce extensively smoothed curves.
- **Spline interpolation**: smoothing using spline interpolation via the `spline()` function. This method interpolates between existing vertices and should be used when the resulting smoothed feature must pass through the vertices of the input feature.

In addition to these smoothing functions, `smoothr` offers functions for filling polygon holes and dropping line and polygon fragments based on a size threshold, as well as densification (i.e. adding additional vertices along curves).

## Setup

```{r packages}
library(terra)
library(sf)
library(units)
library(smoothr)
```

## Example data

This package comes with two simple spatial datasets in `sf` format to test the smoothing algorithms on. `jagged_polygons` contains 9 polygons with sharp corners begging to be smoothed out:

```{r jagged-polygons, echo=FALSE}
par(mar = c(0, 0, 0, 0), oma = c(4, 0, 0, 0), mfrow = c(3, 3))
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA)
}
```

Notice that these polygons have a range of complexities, some have holes, and some are mutlipart polygons. Feature attributes help distinguish between the different types.

```{r jagged-polygons_print, echo=FALSE}
print(jagged_polygons)
```

`jagged_lines` contains 9 polylines with disgustingly crooked edges.

```{r jagged-lines, echo=FALSE}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3)
}
```

Again, there's a range of complexities, some lines form closed loops, and some are multipart.

```{r jagged-lines_print, echo=FALSE}
print(jagged_lines)
```

The final dataset that comes with this package, `jagged_raster`, is a simulated occurrence probability for a species, consisting of a spatially auto-correlated raster layer with values between 0 and 1. It is a 25x25 grid of 100 square kilometer cells in a North American centered Albers Equal Area projection. This raster can be used to experiment with smoothing polygons generated from rasters. It is stored within this package as a GeoTIFF so we need to load it using `rast()` from the `terra` package prior to use.

```{r guass-field, results="hide", dev="png", echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
file <- system.file("extdata", "jagged-raster.tif", package = "smoothr")
jagged_raster <- rast(file)
plot(jagged_raster, col = heat.colors(100), axes = FALSE)
```

## Smoothing methods

Currently, three smoothing methods have been implemented: Chaikin's corner cutting algorithm, Gaussian kernel smoothing, and spline interpolation. All are accessed with the `smooth()` function, and all methods work on spatial lines and polygons in `sf` and `sp` format.

### Chaikin's corner cutting algorithm

Chaikin's corner cutting algorithm smooths by iteratively replacing every point by two new points: one 1/4 of the way to the next point and one 1/4 of the way to the previous point. This method applies a moderate amount of smoothing of sharp corners without extensive generalization, and is a good choice when the desire is to smooth without drastically altering the input features. In addition, this algorithm has the benefit of only requiring a single, easily interpretable parameter: the number of smoothing iterations. 

This method can be applied with `smooth(x, method = "chaikin")`. Here's what this looks like for the polygons:

```{r chaikin-polygons, echo=-1}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
p_smooth_chaikin <- smooth(jagged_polygons, method = "chaikin")
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA)
  plot(st_geometry(p_smooth_chaikin[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
}
```

And for the lines:

```{r chaikin-lines, echo=-1}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
l_smooth_chaikin <- smooth(jagged_lines, method = "chaikin")
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3)
  plot(st_geometry(l_smooth_chaikin[i, ]), col = "#E41A1C", lwd = 2, add = TRUE)
}
```

### Kernel smoothing

This method applies Gaussian kernel regression to the x and y coordinates independently using the built-in `ksmooth()` function. Prior to smoothing, additional vertices are added via densification with `smooth_densify()`. For polygons (and closed lines), `method = "periodic"` is used to avoid getting a kink at the start/end of the curve defining the boundary. 

Kernel smoothing simultaneously smooths and generalizes curves, and can be tuned to produce drastically smoothed curves. This method produces results similar to the [PAEK algorithm used in ArcGIS](http://pro.arcgis.com/en/pro-app/tool-reference/cartography/how-smooth-line-and-smooth-polygon-work.htm). `smooth_ksmooth()` has parameters specifying the number of points in the resulting feature (either the number of sub-segments to split each line segment into or the maximum distance between points) and the bandwidth of the Gaussian kernel. Choosing a suitable bandwidth is critical for correctly smoothing features using this algorithm and typically users will want to explore a range of bandwidth to find a value that works for their particular scenario.

This method can be applied with `smooth(x, method = "ksmooth")`. Here's what this looks like for the polygons:

```{r smooth-polygons, echo=-1}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
p_smooth_ksmooth <- smooth(jagged_polygons, method = "ksmooth")
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA)
  plot(st_geometry(p_smooth_ksmooth[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
}
```

And for the lines:

```{r ksmooth-lines, echo=-1}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
l_smooth_ksmooth <- smooth(jagged_lines, method = "ksmooth")
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3)
  plot(st_geometry(l_smooth_ksmooth[i, ]), col = "#E41A1C", lwd = 2, add = TRUE)
}
```

### Spline interpolation

This method applies a spline interpolation to the x and y coordinates independently using the built-in `spline()` function. For polygons (and closed lines), `method = "periodic"` is used to avoid getting a kink at the start/end of the curve defining the boundary. Unlike the corner cutting algorithm, this method results in a curve that passes through the vertices of the original curve, which may be a desirable feature. Unfortunately, this results in an unnaturally wiggly curve. Spline interpolation requires a parameter specifying the number of points to interpolate at, which can either be an absolute number or a relative increase in the number of vertices.

This method can be applied with `smooth(x, method = "spline")`. Here's what this looks like for the polygons:

```{r spline-polygons, echo=-1}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
p_smooth_spline <- smooth(jagged_polygons, method = "spline")
for (i in seq_len(nrow(jagged_polygons))) {
  plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = NA)
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA,
       add = TRUE)
  plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
}
```

And for the lines:

```{r spline-lines, echo=-1}
par(mar = c(0, 0, 0, 0), oma = c(0, 0, 0, 0), mfrow = c(3, 3))
l_smooth_spline <- smooth(jagged_lines, method = "spline")
for (i in seq_len(nrow(jagged_lines))) {
  plot(st_geometry(l_smooth_spline[i, ]), col = NA)
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3, add = TRUE)
  plot(st_geometry(l_smooth_spline[i, ]), col = "#E41A1C", lwd = 2, add = TRUE)
}
```

## Other goodies

`smoothr` contains some additional tools that aren't strictly for smoothing, but can help clean up polygons and lines and make them appear more aesthetically pleasing. I'll outline these miscellaneous tools below.

### Densification

Densification is the process of adding additional points to the line segments defining polylines or polygons. Each line segment is split into multiple sub-segments of equal length, and the original vertices are always included in the densified feature. Note that the densification algorithm treats all vertices as Euclidean points, i.e. new points will not fall on a great circle between existing vertices, rather they'll be along a straight line.

Use `smooth(method = "densify")`, or it's alias `densify()` to perform densification. The degree of densification can either be specified as the number of sub-segments each line segment is split into (`n`):

```{r densify-n, echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
l <- jagged_lines$geometry[[2]]
# split every segment into 2
l_dense <- densify(l, n = 2)
plot(l, lwd = 5)
plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19, add = TRUE)
```

Or the maximum Euclidean distance between vertices (`max_distance`):

```{r densify-md, echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
l <- jagged_lines$geometry[[2]]
# split every segment into 2
l_dense <- densify(l, max_distance = 0.1)
plot(l, lwd = 5)
plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19, add = TRUE)
```

### Drop crumbs

`drop_crumbs()` removes small lines or polygons based on a length or area threshold. For multipart features, this function dives into the individual component features and applies the threshold. The threshold can either be provided as a number, assumed to be in the units of the coordinates reference system (meters for unprojected coordinates), or as a `units` object. This latter approach provides more flexibility because a threshold can be given in any type of units and it will be converted to the correct units automatically. 

For example, to remove polygons less than 200 square kilometers:

```{r drop-crumbs-polygons, echo=-1}
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
p <- jagged_polygons$geometry[7]
area_thresh <- units::set_units(200, km^2)
p_dropped <- drop_crumbs(p, threshold = area_thresh)
plot(p, col = "black", main = "Original")
plot(p_dropped, col = "black", main = "After drop_crumbs()")
```

And, to remove lines shorter than 25 miles:

```{r drop-crumbs-lines, echo=-1}
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
l <- jagged_lines$geometry[8]
# note that any units can be used
# conversion to units of projection happens automatically
length_thresh <- units::set_units(25, miles)
l_dropped <- drop_crumbs(l, threshold = length_thresh)
plot(l, lwd = 5, main = "Original")
plot(l_dropped, lwd = 5, main = "After drop_crumbs()")
```

### Fill holes

`fill_holes()` fills (i.e. removes) holes in polygons when they are below a given area threshold. As with `drop_crumbs()`, the threshold can either be provided as a number or a `units` object.

For example, to remove holes less than 800 square kilometers in size:

```{r fill-polygons, echo=-1}
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
p <- jagged_polygons$geometry[5]
area_thresh <- units::set_units(800, km^2)
p_dropped <- fill_holes(p, threshold = area_thresh)
# plot
plot(p, col = "black", main = "Original")
plot(p_dropped, col = "black", main = "After fill_holes()")
```

## Raster-to-polygon conversion

The whole point of this `smoothr` business was to smooth out polygons generated from rasters, so let's work through a quick example of that. Treating `jagged_raster` as the occurrence probability for a species, imagine we want to produce a range map for this species, showing where it occurs with at least 50% probability. We can convert the raster to a binary presence/absence map, then polygonize.

```{r polygonize, dev="png", echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
file <- system.file("extdata", "jagged-raster.tif", package = "smoothr")
jagged_raster <- rast(file)

# pres/abs map
r <- classify(jagged_raster, rcl = c(-Inf, 0.5, Inf)) |>
  as.numeric()
plot(r, col = c("white", "#4DAF4A"), legend = FALSE, axes = FALSE)

# polygonize
r_poly <- ifel(r > 0, r, NA) |>
  as.polygons() |>
  st_as_sf()
plot(r_poly, col = NA, border = "grey20", lwd = 1.5, add = TRUE)
```

To start cleaning this up, let's remove all the small polygons resulting from a single raster cell. Recall that the cells is this raster are 100 square kilometers each.

```{r raster-drop, echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_dropped <- drop_crumbs(r_poly, set_units(101, km^2))
# plot
plot(r_poly_dropped, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)
```

There are two holes in the large central polygon. Next, let's fill in the smaller of the two holes, which is two raster cells (200 square kilometers) in size, but keep the larger one.

```{r raster-fill, echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_filled <- fill_holes(r_poly_dropped, set_units(201, km^2))
# plot
plot(r_poly_filled, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)
```

Finally, to make this more aesthetically pleasing, let's smooth out those sharp edges.

```{r smooth-raster, echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_smooth <- smooth(r_poly_filled, method = "ksmooth")
# plot
plot(r_poly_smooth, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)
```

For a greater degree of smoothing and generalization, increase the `smoothness` parameter:

```{r smooth-raster-more, echo=-1}
par(mar = c(0, 0, 0, 0), mfrow = c(1, 1))
r_poly_smooth <- smooth(r_poly_filled, method = "ksmooth", smoothness = 2)
# plot
plot(r_poly_smooth, col = "#4DAF4A", border = "grey20", lwd = 1.5, main = NULL)
```

Nice, so much easier on the eyes! 

## References

Chaikin's corner cutting algorithm:

- Chaikin, G. An algorithm for high speed curve generation. Computer Graphics and Image Processing 3 (1974), 346–349
-http://www.cs.unc.edu/~dm/UNC/COMP258/LECTURES/Chaikins-Algorithm.pdf
- Where to find Python implementation of Chaikin's corner cutting algorithm?: `https://stackoverflow.com/a/47255374/3591386`

Kernel smoothing:

- Nadaraya-Watson Optimal Bandwidth: `https://stats.stackexchange.com/a/143608/44268`

Spline interpolation:

- Create polygon from set of points distributed: `https://stackoverflow.com/questions/26087772/26089377`
- Smoothing polygons in contour map?: `https://gis.stackexchange.com/questions/24827/24929`

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smoothr-package.R
\docType{package}
\name{smoothr-package}
\alias{smoothr}
\alias{smoothr-package}
\title{\code{smoothr}: Smooth and Tidy Spatial Features}
\description{
Tools for smoothing and tidying spatial features (i.e. lines and polygons) to
make them more aesthetically pleasing. Smooth curves, fill holes, and remove
small fragments from lines and polygons.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://strimas.com/smoothr/}
  \item \url{https://github.com/mstrimas/smoothr}
  \item Report bugs at \url{https://github.com/mstrimas/smoothr/issues}
}

}
\author{
\strong{Maintainer}: Matthew Strimas-Mackey \email{mstrimas@gmail.com} (\href{https://orcid.org/0000-0001-8929-7776}{ORCID})

Authors:
\itemize{
  \item Matthew Strimas-Mackey \email{mstrimas@gmail.com} (\href{https://orcid.org/0000-0001-8929-7776}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fill-holes.R
\name{fill_holes}
\alias{fill_holes}
\title{Fill small holes in polygons}
\usage{
fill_holes(x, threshold)
}
\arguments{
\item{x}{spatial features; lines or polygons from either the \code{sf}, \code{sp}, or
\code{terra} packages.}

\item{threshold}{an area threshold, below which holes will be removed.
Provided either as a \code{units} object (see \code{\link[units:set_units]{units::set_units()}}), or a
numeric threshold in the units of the coordinate reference system. If \code{x}
is in unprojected coordinates, a numeric threshold is assumed to be in
square meters. A threshold of 0 will return the input polygons unchanged.}
}
\value{
A spatial feature, with holes filled, in the same format as the input
data.
}
\description{
Fill polygon holes that fall below a given area threshold.
}
\examples{
# fill holes smaller than 1000km2
p <- jagged_polygons$geometry[5]
area_thresh <- units::set_units(1000, km^2)
p_dropped <- fill_holes(p, threshold = area_thresh)
# plot
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
plot(p, col = "black", main = "Original")
plot(p_dropped, col = "black", main = "After fill_holes()")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth-ksmooth.R
\name{smooth_ksmooth}
\alias{smooth_ksmooth}
\title{Kernel smooth}
\usage{
smooth_ksmooth(
  x,
  wrap = FALSE,
  smoothness = 1,
  bandwidth,
  n = 10L,
  max_distance
)
}
\arguments{
\item{x}{numeric matrix; 2-column matrix of coordinates.}

\item{wrap}{logical; whether the coordinates should be wrapped at the ends,
as for polygons and closed lines, to ensure a smooth edge.}

\item{smoothness}{numeric; a parameter controlling the bandwidth of the
Gaussian kernel, and therefore the smoothness and level of generalization.
By default, the bandwidth is chosen as the mean distance between adjacent
points. The \code{smoothness} parameter is a multiplier of this chosen
bandwidth, with values greater than 1 yielding more highly smoothed and
generalized features and values less than 1 yielding less smoothed and
generalized features.}

\item{bandwidth}{numeric; the bandwidth of the Guassian kernel. If this
argument is supplied, then \code{smoothness} is ignored and an optimal bandwidth
is not estimated.}

\item{n}{integer; number of times to split each line segment for
\code{\link[=smooth_densify]{smooth_densify()}}. Ignored if \code{max_distance} is specified.}

\item{max_distance}{numeric; the maximum distance between vertices for
\code{\link[=smooth_densify]{smooth_densify()}}. This is the Euclidean distance and not the great circle
distance.}
}
\value{
A matrix with the coordinates of the smoothed curve.
}
\description{
Kernel smoothing uses \code{\link[stats:ksmooth]{stats::ksmooth()}} to smooth out existing vertices
using Gaussian kernel regression. Kernel smoothing is applied to the \code{x} and
\code{y} coordinates are independently. Prior to smoothing, \code{\link[=smooth_densify]{smooth_densify()}} is
called to generate additional vertices, and the smoothing is applied to this
densified set of vertices.
}
\details{
Kernel smoothing both smooths and generalizes curves, and the extent of these
effects is dependent on the bandwidth of the smoothing kernel. Therefore,
choosing a sensible bandwidth is critical when using this method. The choice
of bandwidth will be dependent on the projection, scale, and desired amount
of smoothing and generalization. The are two methods of adjusting the
bandwidth. By default, the bandwidth will be set to the average distances
between adjacent vertices. The \code{smoothness} factor can then be used to adjust
this calculated bandwidth, values greater than 1 will lead to more smoothing,
values less than 1 will lead to less smoothing. Alternatively, the bandwidth
can be chosen manually with the \code{bandwidth} argument. Typically, users will
need to explore a range of bandwidths to determine which yields the best
results for their situation.

This function works on matrices of points and is generally not called
directly. Instead, use \code{\link[=smooth]{smooth()}} with \code{method = "ksmooth"} to apply this
smoothing algorithm to spatial features.
}
\examples{
# smooth_ksmooth works on matrices of coordinates
# use the matrix of coordinates defining a polygon as an example
m <- jagged_polygons$geometry[[2]][[1]]
m_smooth <- smooth_ksmooth(m, wrap = TRUE)
class(m)
class(m_smooth)
plot(m, type = "l", col = "black", lwd = 3, axes = FALSE, xlab = NA,
     ylab = NA)
lines(m_smooth, lwd = 3, col = "red")

# lines can also be smoothed
l <- jagged_lines$geometry[[2]][]
l_smooth <- smooth_ksmooth(l, wrap = FALSE, max_distance = 0.05)
plot(l, type = "l", col = "black", lwd = 3, axes = FALSE, xlab = NA,
     ylab = NA)
lines(l_smooth, lwd = 3, col = "red")

# explore different levels of smoothness
p <- jagged_polygons$geometry[[2]][[1]]
ps1 <- smooth_ksmooth(p, wrap = TRUE, max_distance = 0.01, smoothness = 0.5)
ps2 <- smooth_ksmooth(p, wrap = TRUE, max_distance = 0.01, smoothness = 1)
ps3 <- smooth_ksmooth(p, wrap = TRUE, max_distance = 0.01, smoothness = 2)
# plot
par(mar = c(0, 0, 0, 0), oma = c(10, 0, 0, 0))
plot(p, type = "l", col = "black", lwd = 3, axes = FALSE, xlab = NA,
     ylab = NA)
lines(ps1, lwd = 3, col = "#E41A1C")
lines(ps2, lwd = 3, col = "#4DAF4A")
lines(ps3, lwd = 3, col = "#377EB8")
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), new = TRUE)
plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n", axes = FALSE)
legend("bottom", legend = c("0.5", "1", "2"),
       col = c("#E41A1C", "#4DAF4A", "#377EB8"),
       lwd = 3, cex = 2, box.lwd = 0, inset = 0, horiz = TRUE)

library(sf)
p <- jagged_polygons$geometry[[2]]
p_smooth <- smooth(p, method = "ksmooth")
class(p)
class(p_smooth)
plot(p_smooth, border = "red")
plot(p, add = TRUE)
}
\references{
The kernel smoothing method was inspired by the following
StackExchange answers:
\itemize{
\item \url{https://stats.stackexchange.com/a/143608/44268}
\item \url{https://gis.stackexchange.com/a/24929/26661}
}
}
\seealso{
\code{\link[=smooth]{smooth()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{jagged_lines_3d}
\alias{jagged_lines_3d}
\title{3D jagged line with Z-dimension for smoothing}
\format{
An \link[sf:sf]{sf} object with 9 features and 3 attributes:
\itemize{
\item \strong{\code{type}}: character; the geometry, i.e. "polygon" or "line".
\item \strong{\code{closed}}: logical; whether the line forms a closed loop or not.
\item \strong{\code{multipart}}: logical; whether the feature is single or multipart.
}
}
\usage{
jagged_lines_3d
}
\description{
Spatial lines in \link[sf:sf]{sf}  format for smoothing in three dimensions.
There are examples of open and closed loops
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth-chaikin.R
\name{smooth_chaikin}
\alias{smooth_chaikin}
\title{Chaikin's corner cutting algorithm}
\usage{
smooth_chaikin(x, wrap = FALSE, refinements = 3L)
}
\arguments{
\item{x}{numeric matrix; 2-column matrix of coordinates.}

\item{wrap}{logical; whether the coordinates should be wrapped at the ends,
as for polygons and closed lines, to ensure a smooth edge.}

\item{refinements}{integer; number of corner cutting iterations to apply.}
}
\value{
A matrix with the coordinates of the smoothed curve.
}
\description{
Chaikin's corner cutting algorithm smooths a curve by iteratively replacing
every point by two new points: one 1/4 of the way to the next point and one
1/4 of the way to the previous point.
}
\details{
This function works on matrices of points and is generally not called
directly. Instead, use \code{\link[=smooth]{smooth()}} with \code{method = "chaikin"} to apply this
smoothing algorithm to spatial features.
}
\examples{
# smooth_chaikin works on matrices of coordinates
# use the matrix of coordinates defining a polygon as an example
m <- jagged_polygons$geometry[[2]][[1]]
m_smooth <- smooth_chaikin(m, wrap = TRUE)
class(m)
class(m_smooth)
plot(m, type = "l", axes = FALSE, xlab = NA, ylab = NA)
lines(m_smooth, col = "red")

# smooth is a wrapper for smooth_chaikin that works on spatial features
library(sf)
p <- jagged_polygons$geometry[[2]]
p_smooth <- smooth(p, method = "chaikin")
class(p)
class(p_smooth)
plot(p)
plot(p_smooth, border = "red", add = TRUE)
}
\references{
The original reference for Chaikin's corner cutting algorithm is:
\itemize{
\item Chaikin, G. An algorithm for high speed curve generation.
\emph{Computer Graphics and Image Processing} 3 (1974), 346-349.
}

This implementation was inspired by the following StackOverflow answer:
\itemize{
\item \url{https://stackoverflow.com/a/47255374/3591386}
}
}
\seealso{
\code{\link[=smooth]{smooth()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth.R
\name{smooth}
\alias{smooth}
\title{Smooth a spatial feature}
\usage{
smooth(x, method = c("chaikin", "ksmooth", "spline", "densify"), ...)
}
\arguments{
\item{x}{spatial features; lines or polygons from either the \code{sf}, \code{sp}, or
\code{terra} packages.}

\item{method}{character; specifies the type of smoothing method to use.
Possible methods are: \code{"chaikin"}, \code{"ksmooth"}, \code{"spline"}, and
\code{"densify"}. Each method has one or more parameters specifying the amount
of smoothing to perform. See Details for descriptions.}

\item{...}{additional arguments specifying the amount of smoothing, passed on
to the specific smoothing function, see Details below.}
}
\value{
A smoothed polygon or line in the same format as the input data.
}
\description{
Smooth out the jagged or sharp corners of spatial lines or polygons to make
them appear more aesthetically pleasing and natural.
}
\details{
Specifying a method calls one of the following underlying smoothing
functions. Each smoothing method has one or more parameters that specify
the extent of smoothing. Note that for multiple features, or multipart
features, these parameters apply to each individual, singlepart feature.
\itemize{
\item \code{\link[=smooth_chaikin]{smooth_chaikin()}}: Chaikin's corner cutting algorithm smooths a curve by
iteratively replacing every point by two new points: one 1/4 of the way to
the next point and one 1/4 of the way to the previous point. Smoothing
parameters:
\itemize{
\item \code{refinements}: number of corner cutting iterations to apply.
}
\item \code{\link[=smooth_ksmooth]{smooth_ksmooth()}}: kernel smoothing via the \code{\link[stats:ksmooth]{stats::ksmooth()}} function.
This method first calls \code{\link[=smooth_densify]{smooth_densify()}} to densify the feature, then
applies Gaussian kernel regression to smooth the resulting points.
Smoothing parameters:
\itemize{
\item \code{smoothness}: a positive number controlling the smoothness and level of
generalization. At the default value of 1, the  bandwidth is chosen as
the mean distance between adjacent vertices. Values greater than 1
increase the bandwidth, yielding more highly smoothed and generalized
features, and values less than 1 decrease the bandwidth, yielding less
smoothed and generalized features.
\item \code{bandwidth}: the bandwidth of the Guassian kernel. If this argument is
supplied, then \code{smoothness} is ignored and an optimal bandwidth is not
estimated.
\item \code{n}: number of times to split each line segment in the densification
step. Ignored if \code{max_distance} is specified.
\item \code{max_distance}: the maximum distance between vertices in the resulting
features for the densification step. This is the Euclidean distance and
not the great circle distance.
}
\item \code{\link[=smooth_spline]{smooth_spline()}}: spline interpolation via the \code{\link[stats:spline]{stats::spline()}}
function. This method interpolates between existing vertices and can be
used when the resulting smoothed feature should pass through the vertices
of the input feature. Smoothing parameters:
\itemize{
\item \code{vertex_factor}: the proportional increase in the number of vertices in
the smooth feature. For example, if the original feature has 100
vertices, a value of 2.5 will yield a new, smoothed feature with 250
vertices.  Ignored if \code{n} is specified.
\item \code{n}: number of vertices in each smoothed feature.
}
\item \code{\link[=smooth_densify]{smooth_densify()}}: densification of vertices for lines and polygons.
This is not a true smoothing algorithm, rather new vertices are added to
each line segment via linear interpolation. Densification parameters:
\itemize{
\item \code{n}: number of times to split each line segment. Ignored if
\code{max_distance} is specified.
\item \code{max_distance}: the maximum distance between vertices in the resulting
feature. This is the Euclidean distance and not the great circle
distance.
}
}
}
\examples{
library(sf)
# compare different smoothing methods
# polygons
par(mar = c(0, 0, 0, 0), oma = c(4, 0, 0, 0), mfrow = c(3, 3))
p_smooth_chaikin <- smooth(jagged_polygons, method = "chaikin")
p_smooth_ksmooth <- smooth(jagged_polygons, method = "ksmooth")
p_smooth_spline <- smooth(jagged_polygons, method = "spline")
for (i in 1:nrow(jagged_polygons)) {
  plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = NA)
  plot(st_geometry(jagged_polygons[i, ]), col = "grey40", border = NA,
       add = TRUE)
  plot(st_geometry(p_smooth_chaikin[i, ]), col = NA, border = "#E41A1C",
       lwd = 2, add = TRUE)
  plot(st_geometry(p_smooth_ksmooth[i, ]), col = NA, border = "#4DAF4A",
       lwd = 2, add = TRUE)
  plot(st_geometry(p_smooth_spline[i, ]), col = NA, border = "#377EB8",
       lwd = 2, add = TRUE)
}
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), new = TRUE)
plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n", axes = FALSE)
legend("bottom", legend = c("chaikin", "ksmooth", "spline"),
       col = c("#E41A1C", "#4DAF4A", "#377EB8"),
       lwd = 2, cex = 2, box.lwd = 0, inset = 0, horiz = TRUE)

# lines
par(mar = c(0, 0, 0, 0), oma = c(4, 0, 0, 0), mfrow = c(3, 3))
l_smooth_chaikin <- smooth(jagged_lines, method = "chaikin")
l_smooth_ksmooth <- smooth(jagged_lines, method = "ksmooth")
l_smooth_spline <- smooth(jagged_lines, method = "spline")
for (i in 1:nrow(jagged_lines)) {
  plot(st_geometry(l_smooth_spline[i, ]), col = NA)
  plot(st_geometry(jagged_lines[i, ]), col = "grey20", lwd = 3, add = TRUE)
  plot(st_geometry(l_smooth_chaikin[i, ]), col = "#E41A1C", lwd = 2,
       lty = 2, add = TRUE)
  plot(st_geometry(l_smooth_ksmooth[i, ]), col = "#4DAF4A", lwd = 2,
       lty = 2, add = TRUE)
  plot(st_geometry(l_smooth_spline[i, ]), col = "#377EB8", lwd = 2,
       lty = 2, add = TRUE)
}
par(fig = c(0, 1, 0, 1), oma = c(0, 0, 0, 0), new = TRUE)
plot(0, 0, type = "n", bty = "n", xaxt = "n", yaxt = "n", axes = FALSE)
legend("bottom", legend = c("chaikin", "ksmooth", "spline"),
       col = c("#E41A1C", "#4DAF4A", "#377EB8"),
       lwd = 2, cex = 2, box.lwd = 0, inset = 0, horiz = TRUE)
}
\references{
See specific smoothing function help pages for references.
}
\seealso{
\code{\link[=smooth_chaikin]{smooth_chaikin()}} \code{\link[=smooth_ksmooth]{smooth_ksmooth()}} \code{\link[=smooth_spline]{smooth_spline()}}
\code{\link[=smooth_densify]{smooth_densify()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop-crumbs.R
\name{drop_crumbs}
\alias{drop_crumbs}
\title{Remove small polygons or line segments}
\usage{
drop_crumbs(x, threshold, drop_empty = TRUE)
}
\arguments{
\item{x}{spatial features; lines or polygons from either the \code{sf}, \code{sp}, or
\code{terra} packages.}

\item{threshold}{an area or length threshold, below which features will be
removed. Provided either as a \code{units} object (see \code{\link[units:set_units]{units::set_units()}}), or
a numeric threshold in the units of the coordinate reference system. If \code{x}
is in unprojected coordinates, a numeric threshold is assumed to be in
meters.}

\item{drop_empty}{logical; whether features with sizes below the given
threshold should be removed (the default) or kept as empty geometries. Note
that \code{sp} objects cannot store empty geometries, so this argument will be
ignored and empty geometries will always be removed.}
}
\value{
A spatial feature, with small pieces removed, in the same format as
the input data. If none of the features are larger than the threshold,
\code{sf} inputs will return a geometry set with zero features, and \code{sp} inputs
will return \code{NULL}.
}
\description{
Remove polygons or line segments below a given area or length threshold.
}
\details{
For multipart features, the removal threshold is applied to the individual
components. This means that, in some cases, an entire feature may be removed,
while in other cases, only parts of the multipart feature will be removed.
}
\examples{
# remove polygons smaller than 200km2
p <- jagged_polygons$geometry[7]
area_thresh <- units::set_units(200, km^2)
p_dropped <- drop_crumbs(p, threshold = area_thresh)
# plot
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
plot(p, col = "black", main = "Original")
if (length(p_dropped) > 0) {
  plot(p_dropped, col = "black", main = "After drop_crumbs()")
}


# remove lines less than 25 miles
l <- jagged_lines$geometry[8]
# note that any units can be used
# conversion to units of projection happens automatically
length_thresh <- units::set_units(25, miles)
l_dropped <- drop_crumbs(l, threshold = length_thresh)
# plot
par(mar = c(0, 0, 1, 0), mfrow = c(1, 2))
plot(l, lwd = 5, main = "Original")
if (length(l_dropped)) {
  plot(l_dropped, lwd = 5, main = "After drop_crumbs()")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{jagged_polygons}
\alias{jagged_polygons}
\title{Jagged polygons for smoothing}
\format{
An \link[sf:sf]{sf} object with 9 features and 3 attributes:
\itemize{
\item \strong{\code{type}}: character; the geometry, i.e. "polygon" or "line".
\item \strong{\code{hole}}: logical; whether the polygon has holes or not.
\item \strong{\code{multipart}}: logical; whether the feature is single or multipart.
}
}
\usage{
jagged_polygons
}
\description{
Spatial polygons in \link[sf:sf]{sf} format for smoothing. Most of these polygons
have been created by converting rasters to polygons and therefore consist
entirely of right angles. There are examples of polygons with holes and
multipart polygons.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth-spline.R
\name{smooth_spline}
\alias{smooth_spline}
\title{Spline interpolation}
\usage{
smooth_spline(x, wrap = FALSE, vertex_factor = 5, n)
}
\arguments{
\item{x}{numeric matrix; matrix of coordinates.}

\item{wrap}{logical; whether the coordinates should be wrapped at the ends,
as for polygons and closed lines, to ensure a smooth edge.}

\item{vertex_factor}{double; the proportional increase in the number of
vertices in the smooth curve. For example, if the original curve has 100
points, a value of \code{2.5} will yield a new smoothed curve with 250 points.
Ignored if \code{n} is specified.}

\item{n}{integer; number of vertices in the smoothed curve.}
}
\value{
A matrix with the coordinates of the smoothed curve.
}
\description{
Spline interpolation uses \code{\link[stats:spline]{stats::spline()}} to interpolate between existing
vertices using piecewise cubic polynomials. The coordinates are interpolated
independently. The curve will always pass through the vertices of the
original feature.
}
\details{
This function works on matrices of points and is generally not called
directly. Instead, use \code{\link[=smooth]{smooth()}} with \code{method = "spline"} to apply this
smoothing algorithm to spatial features.
}
\examples{
# smooth_spline works on matrices of coordinates
# use the matrix of coordinates defining a polygon as an example
m <- jagged_polygons$geometry[[2]][[1]]
m_smooth <- smooth_spline(m, wrap = TRUE)
class(m)
class(m_smooth)
plot(m_smooth, type = "l", col = "red", axes = FALSE, xlab = NA, ylab = NA)
lines(m, col = "black")

# smooth is a wrapper for smooth_spline that works on spatial features
library(sf)
p <- jagged_polygons$geometry[[2]]
p_smooth <- smooth(p, method = "spline")
class(p)
class(p_smooth)
plot(p_smooth, border = "red")
plot(p, add = TRUE)
}
\references{
The spline method was inspired by the following StackExchange
answers:
\itemize{
\item \url{https://stackoverflow.com/a/26089377/3591386}
\item \url{https://gis.stackexchange.com/a/24929/26661}
}
}
\seealso{
\code{\link[=smooth]{smooth()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth-densify.R
\name{smooth_densify}
\alias{smooth_densify}
\title{Densify lines or polygons}
\usage{
smooth_densify(x, wrap = FALSE, n = 10L, max_distance)
}
\arguments{
\item{x}{numeric matrix; matrix of coordinates.}

\item{wrap}{logical; whether the coordinates should be wrapped at the ends,
as for polygons and closed lines, to ensure a smooth edge.}

\item{n}{integer; number of times to split each line segment. Ignored if
\code{max_distance} is specified.}

\item{max_distance}{numeric; the maximum distance between vertices in the
resulting matrix. This is the Euclidean distance and not the great circle
distance.}
}
\value{
A matrix with the coordinates of the densified curve.
}
\description{
This function adds additional vertices to lines or polygons via linear
interpolation, always while keeping the original vertices. Each line segment
will be split into equal length sub-segments. This densification algorithm
treats all vertices as Euclidean points, i.e. new points will not fall on a
great circle between existing vertices, rather they'll be along a straight
line.
}
\details{
This function works on matrices of points and is generally not called
directly. Instead, use \code{\link[=smooth]{smooth()}} with \code{method = "densify"} to apply this
smoothing algorithm to spatial features.
}
\examples{
# smooth_densify works on matrices of coordinates
# use the matrix of coordinates defining a line as an example
m <- jagged_lines$geometry[[2]][]
m_dense <- smooth_densify(m, n = 5)
class(m)
class(m_dense)
plot(m, type = "b", pch = 19, cex = 1.5, axes = FALSE, xlab = NA, ylab = NA)
points(m_dense, col = "red", pch = 19, cex = 0.5)

# max_distance can be used to ensure vertices are at most a given dist apart
m_md <- smooth_densify(m, max_distance = 0.05)
plot(m, type = "b", pch = 19, cex = 1.5, axes = FALSE, xlab = NA, ylab = NA)
points(m_md, col = "red", pch = 19, cex = 0.5)

# smooth is a wrapper for smooth_densify that works on spatial features
library(sf)
l <- jagged_lines$geometry[[2]]
l_dense <- smooth(l, method = "densify", n = 2)
class(l)
class(l_dense)
plot(l, lwd = 5)
plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19,
     add = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/densify.R
\name{densify}
\alias{densify}
\title{Densify spatial lines or polygons}
\usage{
densify(x, n = 10L, max_distance)
}
\arguments{
\item{x}{spatial features; lines or polygons from either the \code{sf}, \code{sp}, or
\code{terra} packages.}

\item{n}{integer; number of times to split each line segment. Ignored if
\code{max_distance} is specified.}

\item{max_distance}{numeric; the maximum distance between vertices in the
resulting matrix. This is the Euclidean distance and not the great circle
distance.}
}
\value{
A densified polygon or line in the same format as the input data.
}
\description{
A wrapper for \code{smooth(x, method = "densify")}. This function adds additional
vertices to spatial feature via linear interpolation, always while keeping
the original vertices. Each line segment will be split into equal length
sub-segments. This densification algorithm treats all vertices as Euclidean
points, i.e. new points will not fall on a great circle between existing
vertices, rather they'll be along a straight line.
}
\examples{
library(sf)
l <- jagged_lines$geometry[[2]]
l_dense <- densify(l, n = 2)
plot(l, lwd = 5)
plot(l_dense, col = "red", lwd = 2, lty = 2, add = TRUE)
plot(l_dense |> st_cast("MULTIPOINT"), col = "red", pch = 19,
     add = TRUE)
}
